# -*- coding: utf-8 -*-


"""
The `ImageResources` section.

Image resource blocks are the basic building unit of several file
formats, including Photoshop's native file format, JPEG, and
TIFF. Image resources are used to store non-pixel data associated with
images, such as pen tool paths.
"""


from __future__ import unicode_literals, absolute_import


import struct


import six


import numpy as np
import traitlets as t


from . import docs
from . import enums
from . import util


class _ImageResourceBlockMeta(type(t.HasTraits)):
    """
    A metaclass that builds a mapping of subclasses.
    """
    mapping = {}

    def __new__(cls, name, parents, dct):
        new_cls = type(t.HasTraits).__new__(cls, name, parents, dct)

        if 'resource_id' in dct and isinstance(dct['resource_id'], int):
            if dct['resource_id'] in cls.mapping:
                raise ValueError(
                    "Duplicate resource_id '{}'".format(
                        dct['resource_id']))
            cls.mapping[dct['resource_id']] = new_cls

        return new_cls


@six.add_metaclass(_ImageResourceBlockMeta)
class ImageResourceBlock(t.HasTraits):
    """
    Stores a single image resource block.

    ``pytoshop`` currently doesn't deeply parse image resource
    blocks.  The raw data is merely retained for round-tripping.
    """
    name = t.Unicode(
        default_value='',
        max=255,
        help="Name of image resource."
    )

    def length(self, header):
        data_length = self.data_length(header)
        length = (
            4 + 2 +
            util.pascal_string_length(self.name, 2) +
            4 + data_length
        )
        if data_length % 2 != 0:
            length += 1
        return length
    length.__doc__ = docs.length

    def total_length(self, header):
        return self.length(header)
    total_length.__doc__ = docs.total_length

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        signature = fd.read(4)
        if signature != b'8BIM':
            raise ValueError('Invalid image resource block signature')

        resource_id = util.read_value(fd, 'H')
        name = util.read_pascal_string(fd, 2)

        data_length = util.read_value(fd, 'I')

        util.log(
            "resource_id: {}, name: {}, data_length: {}",
            resource_id, name, data_length
        )

        new_cls = _ImageResourceBlockMeta.mapping.get(
            resource_id, GenericImageResourceBlock)
        start = fd.tell()
        result = new_cls.read_data(fd, resource_id, name, data_length, header)
        end = fd.tell()
        if end - start != data_length:
            raise ValueError("{} read the wrong amount".format(new_cls))

        if data_length % 2 != 0:
            fd.read(1)

        return result
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        fd.write(b'8BIM')
        util.write_value(fd, 'H', self.resource_id)
        util.write_pascal_string(fd, self.name, 2)
        length = self.data_length(header)
        util.write_value(fd, 'I', length)
        start = fd.tell()
        self.write_data(fd, header)
        end = fd.tell()
        if end - start != length:
            raise ValueError(
                "{} wrote the wrong amount".format(self.__class__))
        if length % 2 != 0:
            fd.write(b'\0')
    write.__doc__ = docs.write


class GenericImageResourceBlock(ImageResourceBlock):
    resource_id = t.Int(
        help="Type of image resource."
    )
    data = t.Bytes(
        max=(1 << 32),
        help="Raw data of image resource."
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        data = fd.read(length)
        return cls(resource_id=resource_id, name=name, data=data)

    def data_length(self, header):
        return len(self.data)

    @util.trace_write
    def write_data(self, fd, header):
        fd.write(self.data)


class ImageResourceUnicodeString(ImageResourceBlock):
    value = t.Unicode(
        help="Value"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        data = fd.read(length)
        value = util.decode_unicode_string(data)
        return cls(
            name=name, value=value
        )

    def data_length(self, header):
        return util.unicode_string_length(self.value)

    @util.trace_write
    def write_data(self, fd, header):
        fd.write(util.encode_unicode_string(self.value))


class LayersGroupInfo(ImageResourceBlock):
    """
    Layers group information.

    Indicates which layers are locked together.
    """
    resource_id = enums.ImageResourceID.layers_group_info
    group_ids = t.List(
        t.Int,
        help="Layer group ids"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        data = fd.read(length)
        group_ids = np.frombuffer(data, '>u2').tolist()
        return cls(name=name, group_ids=group_ids)

    def data_length(self, header):
        return len(self.group_ids * 2)

    @util.trace_write
    def write_data(self, fd, header):
        data = np.array(self.group_ids, '>u2').tobytes()
        fd.write(data)


class BorderInfo(ImageResourceBlock):
    """
    Border information.
    """
    resource_id = enums.ImageResourceID.border_info
    border_width_num = t.Int(
        min=0, max=65535,
        help="Border width numerator"
    )
    border_width_den = t.Int(
        min=0, max=65535,
        help="Border width denominator")
    unit = t.Enum(
        list(enums.Units),
        default_value=enums.Units.inches,
        help="Unit. See `enums.Units`."
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        num, den, unit = struct.unpack('>HHH', fd.read(6))
        return cls(
            name=name, border_width_num=num, border_width_den=den,
            unit=unit)

    def data_length(self, header):
        return 6

    @util.trace_write
    def write_data(self, fd, header):
        data = struct.pack('>HHH', self.border_width_num,
                           self.border_width_den, self.unit)
        fd.write(data)


class BackgroundColor(ImageResourceBlock):
    """
    Background color.
    """
    resource_id = enums.ImageResourceID.background_color
    color_space = t.Enum(
        list(enums.ColorSpace),
        default_value=enums.ColorSpace.rgb,
        help="The color space. See `enums.ColorSpace`"
    )
    color = t.List(
        t.Int(min=-32767, max=65536), min=1, max=4,
        help="The color data.  If the color data does not require 4 "
        "values, the extra values are undefined and should be included "
        "as zeros."
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        space_id, a, b, c, d = struct.unpack('>HHHHH', fd.read(10))
        if space_id == enums.ColorSpace.lab:
            b -= 32767
            c -= 32767
        return cls(
            name=name, color_space=space_id, color=[a, b, c, d]
        )

    def data_length(self, header):
        return 10

    @util.trace_write
    def write_data(self, fd, header):
        color = self.color[:]
        color.extend([0] * (4 - len(color)))
        a, b, c, d = color
        if self.color_space == enums.ColorSpace.lab:
            b += 32767
            c += 32767
        data = struct.pack(
            '>HHHHH', self.color_space, a, b, c, d
        )
        fd.write(data)


class PrintFlags(ImageResourceBlock):
    """
    Print flags.
    """
    resource_id = enums.ImageResourceID.print_flags
    labels = t.Bool(
        help="labels"
    )
    crop_marks = t.Bool(
        help="crop marks"
    )
    color_bars = t.Bool(
        help="color bars"
    )
    registration_marks = t.Bool(
        help="registration marks"
    )
    negative = t.Bool(
        help="negative"
    )
    flip = t.Bool(
        help="flip"
    )
    interpolate = t.Bool(
        help="interpolate"
    )
    caption = t.Bool(
        help="caption"
    )
    print_flags = t.Bool(
        help="print flags"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        vals = struct.unpack('>BBBBBBBBB', fd.read(9))
        vals = [bool(x) for x in vals]
        return cls(
            name=name, labels=vals[0], crop_marks=vals[1],
            color_bars=vals[2], registration_marks=vals[3],
            negative=vals[4], flip=vals[5], interpolate=vals[6],
            caption=vals[7], print_flags=vals[8]
        )

    def data_length(self, header):
        return 9

    @util.trace_write
    def write_data(self, fd, header):
        vals = [
            self.labels, self.crop_marks, self.color_bars,
            self.registration_marks, self.negative, self.flip,
            self.interpolate, self.caption, self.print_flags
        ]
        vals = [(x and 255 or 0) for x in vals]
        data = struct.pack('>BBBBBBBBB', *vals)
        fd.write(data)


class GuideResourceBlock(t.HasTraits):
    location = t.Int(
        help="Location of guide in document coordinates."
    )
    direction = t.Enum(
        list(enums.GuideDirection),
        default_value=enums.Units.inches,
        help="Guide direction. See `enums.GuideDirection`."
    )

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        data = fd.read(5)
        location, direction = struct.unpack('>IB', data)
        return cls(location=location, direction=direction)

    @util.trace_write
    def write(self, fd, header):
        data = struct.pack('>IB', self.location, self.direction)
        fd.write(data)


class GridAndGuidesInfo(ImageResourceBlock):
    """
    Grid and guides resource.
    """
    resource_id = enums.ImageResourceID.grid_and_guides_info
    version = 1
    grid_hori = t.Int(
        help="Document-specific grid (horizontal). In 1/32 pt."
    )
    grid_vert = t.Int(
        help="Document-specific grid (vertical). In 1/32 pt."
    )
    guides = t.List(
        t.Instance(GuideResourceBlock),
        help="Guides.  See `GuideResourceBlock`."
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        version, grid_hori, grid_vert, nguides = struct.unpack(
            '>IIII', fd.read(16))
        if version != 1:
            raise ValueError(
                "Unknown version {} in grid and guides info block.".format(
                    version))
        guides = []
        for i in range(nguides):
            guides.append(GuideResourceBlock.read(fd, header))
        return cls(
            name=name, grid_hori=grid_hori, grid_vert=grid_vert,
            guides=guides
        )

    def data_length(self, header):
        return 16 + (5 * len(self.guides))

    @util.trace_write
    def write_data(self, fd, header):
        data = struct.pack(
            '>IIII', 1, self.grid_hori, self.grid_vert, len(self.guides)
        )
        fd.write(data)
        for guide in self.guides:
            guide.write(fd, header)


class CopyrightFlag(ImageResourceBlock):
    resource_id = enums.ImageResourceID.copyright_flag
    copyright = t.Bool(
        help="Is copyrighted?"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        copyright = bool(util.read_value(fd, 'B'))
        return cls(
            name=name, copyright=copyright
        )

    def data_length(self, header):
        return 1

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'B', self.copyright and 255 or 0)


class Url(ImageResourceBlock):
    resource_id = enums.ImageResourceID.url
    url = t.Bytes(
        help='URL'
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        url = fd.read(length)
        return cls(
            name=name, url=url
        )

    def data_length(self, header):
        return len(self.url)

    @util.trace_write
    def write_data(self, fd, header):
        fd.write(self.url)


class GlobalAngle(ImageResourceBlock):
    resource_id = enums.ImageResourceID.global_angle
    angle = t.Int(
        min=-360, max=360,
        help="Global light angle for the effect layer"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        angle = util.read_value(fd, 'i')
        return cls(
            name=name, angle=angle
        )

    def data_length(self, header):
        return 4

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'i', self.angle)


class EffectsVisible(ImageResourceBlock):
    resource_id = enums.ImageResourceID.effects_visible
    visible = t.Bool(
        help="Are effects visible?"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        visible = bool(util.read_value(fd, 'B'))
        return cls(
            name=name, visible=visible
        )

    def data_length(self, header):
        return 1

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'B', self.visible and 255 or 0)


class DocumentSpecificIdsSeedNumber(ImageResourceBlock):
    resource_id = enums.ImageResourceID.document_specific_ids_seed_number
    base_value = t.Int(
        help="Base value"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        base_value = bool(util.read_value(fd, 'I'))
        return cls(
            name=name, base_value=base_value
        )

    def data_length(self, header):
        return 4

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'I', self.base_value)


class UnicodeAlphaNames(ImageResourceUnicodeString):
    resource_id = enums.ImageResourceID.unicode_alpha_names


class GlobalAltitude(ImageResourceBlock):
    resource_id = enums.ImageResourceID.global_altitude
    altitude = t.Int(
        help="Global altitude"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        altitude = util.read_value(fd, 'I')
        return cls(
            name=name, altitude=altitude
        )

    def data_length(self, header):
        return 4

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'I', self.altitude)


class WorkflowUrl(ImageResourceUnicodeString):
    resource_id = enums.ImageResourceID.workflow_url


class AlphaIdentifiers(ImageResourceBlock):
    resource_id = enums.ImageResourceID.alpha_identifiers
    identifiers = t.List(
        t.Int,
        help="Alpha indentifiers"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        length = util.read_value(fd, 'I')
        identifiers = []
        for i in range(length):
            identifiers.append(util.read_value(fd, 'I'))
        return cls(
            name=name, identifiers=identifiers
        )

    def data_length(self, header):
        return 4 + (len(self.identifiers) * 4)

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'I', len(self.identifiers))
        for identifier in self.identifiers:
            util.write_value(fd, 'I', identifier)


class VersionInfo(ImageResourceBlock):
    resource_id = enums.ImageResourceID.version_info
    version = t.Int(
        help="version"
    )
    has_real_merged_data = t.Bool(
        help="has real merged data?"
    )
    writer = t.Unicode(
        help="writer name"
    )
    reader = t.Unicode(
        help="reader name"
    )
    file_version = t.Int(
        help="file version"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        version = util.read_value(fd, 'I')
        has_real_merged_data = bool(util.read_value(fd, 'B'))
        writer = util.read_unicode_string(fd)
        reader = util.read_unicode_string(fd)
        file_version = util.read_value(fd, 'I')
        return cls(
            name=name, version=version,
            has_real_merged_data=has_real_merged_data, writer=writer,
            reader=reader, file_version=file_version
        )

    def data_length(self, header):
        return (
            4 + 1 +
            util.unicode_string_length(self.writer) +
            util.unicode_string_length(self.reader) +
            4)

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'I', self.version)
        util.write_value(fd, 'B', self.has_real_merged_data)
        util.write_unicode_string(fd, self.writer)
        util.write_unicode_string(fd, self.reader)
        util.write_value(fd, 'I', self.file_version)


class PrintScale(ImageResourceBlock):
    resource_id = enums.ImageResourceID.print_scale
    style = t.Enum(
        list(enums.PrintScaleStyle),
        default_value=enums.PrintScaleStyle.centered,
        help="style"
    )
    x = t.Float(
        help="x location"
    )
    y = t.Float(
        help="y location"
    )
    scale = t.Float(
        help="scale"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, resource_id, name, length, header):
        style = util.read_value(fd, 'H')
        x = util.read_value(fd, 'f')
        y = util.read_value(fd, 'f')
        scale = util.read_value(fd, 'f')
        return cls(
            name=name, style=style, x=x, y=y, scale=scale
        )

    def data_length(self, header):
        return (2 + 4 + 4 + 4)

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'H', self.style)
        util.write_value(fd, 'f', self.x)
        util.write_value(fd, 'f', self.y)
        util.write_value(fd, 'f', self.scale)


class ImageResources(t.HasTraits):
    """
    The image resource block section.
    """
    blocks = t.List(
        t.Instance(ImageResourceBlock),
        help="List of all `ImageResourceBlock` items."
    )

    def length(self, header):
        return sum(block.total_length(header) for block in self.blocks)
    length.__doc__ = docs.length

    def total_length(self, header):
        return 4 + self.length(header)
    total_length.__doc__ = docs.total_length

    def get_block(self, resource_id):
        """
        Get the first block with the given resource id.
        """
        for block in self.blocks:
            if block.resource_id == resource_id:
                return block
        return None

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        length = util.read_value(fd, 'I')
        end = fd.tell() + length

        util.log("length: {}, end: {}", length, end)

        blocks = []
        while fd.tell() < end:
            blocks.append(ImageResourceBlock.read(fd, header))

        if fd.tell() != end:
            raise ValueError(
                "read the wrong amount reading image resource blocks")

        return cls(blocks=blocks)
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        util.write_value(fd, 'I', self.length(header))
        for block in self.blocks:
            block.write(fd, header)
    write.__doc__ = docs.write



# -*- coding: utf-8 -*-


"""
Miscellaneous utilities.
"""


from __future__ import unicode_literals, absolute_import


from functools import wraps
import math
import struct
import sys


DEBUG = False


def read_value(fd, fmt, endian='>'):
    """
    Read a single binary value from a file-like object.

    Parameters
    ----------
    fd : file-like object
        Must be opened for reading, in binary mode.

    fmt : str
        A `struct` module `format character
        <https://docs.python.org/2/library/struct.html#format-characters>`__.

    endian : str
        The endianness. Must be ``>`` or ``<``.  Default: ``>``.

    Returns
    -------
    value : any
        The value read from the file.
    """
    fmt = endian + fmt
    size = struct.calcsize(fmt)
    return struct.unpack(fmt, fd.read(size))[0]


def write_value(fd, fmt, value, endian='>'):
    """
    Write a single binary value to a file-like object.

    Parameters
    ----------
    fd : file-like object
        Must be opened for writing, in binary mode.

    fmt : str
        A `struct` module `format character
        <https://docs.python.org/2/library/struct.html#format-characters>`__.

    value : any
        The value to encode and write to the file.

    endian : str
        The endianness. Must be ``>`` or ``<``.  Default: ``>``.
    """
    fmt = endian + fmt
    fd.write(struct.pack(fmt, value))


def pad(number, divisor):
    """
    Pads an integer up to the given divisor.
    """
    if number % divisor:
        number = (number // divisor + 1) * divisor
    return number


def read_pascal_string(fd, padding=1):
    """
    Read a UTF-8-encoded Pascal string from a file.

    Parameters
    ----------
    fd : file-like object
        Must be opened for reading, seekable and in binary mode.

    padding : int, optional
        If provided, additional pad bytes will be read until
        the total amount read is a multiple of padding.

    Returns
    -------
    value : str
        The unicode value of the string.
    """
    length = read_value(fd, 'B')
    if length == 0:
        fd.seek(padding - 1, 1)
        return b''

    result = fd.read(length)

    padded_length = pad(length + 1, padding) - 1
    fd.seek(padded_length - length, 1)
    return result.decode('utf8', 'replace')


def write_pascal_string(fd, value, padding=1):
    """
    Write a UTF-8-encoded Pascal string to a file.

    Parameters
    ----------
    fd : file-like object
        Must be opened for writing and in binary mode.

    value : str
        A unicode string value.

    padding : int, optional
        If provided, additional pad bytes will be written until
        the total amount written is a multiple of padding.
    """
    value = value.encode('utf8')

    length = len(value)
    write_value(fd, 'B', len(value))
    if len(value) == 0:
        fd.write(b'\0' * (padding - 1))
        return

    fd.write(value)

    padding = pad(length + 1, padding) - 1 - length
    if padding != 0:
        fd.write(b'\0' * padding)


def pascal_string_length(value, padding=1):
    """
    Calculates the total length of writing a UTF-8-encoded Pascal
    string to disk.

    Parameters
    ----------
    value : str
        A unicode string value.

    Returns
    -------
    length : int
        The length, in bytes.
    """
    value = value.encode('utf8')

    if len(value) == 0:
        return padding

    length = len(value)
    padding = pad(length + 1, padding) - 1 - length
    return length + padding + 1


def decode_unicode_string(data):
    """
    Decode Photoshop's definition of a `Unicode String
    <https://www.adobe.com/devnet-apps/photoshop/fileformatashtml/#UnicodeStringDefine>`__.
    """
    return data[4:].rstrip(b'\0').decode('utf_16_be')


def encode_unicode_string(s):
    """
    Encode Photoshop's definition of a `Unicode String
    <https://www.adobe.com/devnet-apps/photoshop/fileformatashtml/#UnicodeStringDefine>`__.
    """
    return struct.pack('>L', len(s)) + s.encode('utf_16_be')


def read_unicode_string(fd):
    """
    Read a UTF-16-BE-encoded Unicode string (with length) from a file.

    Parameters
    ----------
    fd : file-like object
        Must be opened for reading, seekable and in binary mode.

    Returns
    -------
    value : str
        The unicode value of the string.
    """
    length = read_value(fd, 'L')
    data = fd.read(length * 2)
    return data.rstrip(b'\0').decode('utf_16_be')


def write_unicode_string(fd, value):
    """
    Write a UTF-16-BE-encoded Unicode string (with length) to a file.

    Parameters
    ----------
    fd : file-like object
        Must be opened for writing and in binary mode.

    value : str
        A unicode string value.
    """
    fd.write(encode_unicode_string(value))


def unicode_string_length(value):
    """
    Calculates the total length of writing a UTF-16-BE-encoded Unicode
    string (with length) to a file.

    Parameters
    ----------
    value : str
        A unicode string value.

    Returns
    -------
    length : int
        The length, in bytes.
    """
    return len(encode_unicode_string(value))


def round_up(x, base=2):
    return int(base * math.ceil(float(x) / base))


_indent = [0]


def trace_read(func):
    """
    Prints debugging information from a read or write method.

    For internal use only.
    """
    @wraps(func)
    def wrapper(self, fd, *args):
        if isinstance(self, type):
            name = self.__name__
        else:
            name = self.__class__.__name__
        log('>>> {} @ {}', name, fd.tell())
        _indent[0] += 1
        result = func(self, fd, *args)
        _indent[0] -= 1
        log('<<< {} @ {}', name, fd.tell())
        return result

    return wrapper


trace_write = trace_read


def log(msg, *args):  # pragma: no cover
    """
    Print a logging message if debugging is turned on.
    """
    if DEBUG:
        print("  " * _indent[0], msg.format(*args))


def is_set_to_default(obj):
    """
    Returns ``True`` if a ``traitlets.HasTraits`` instance is set
    entirely to its defaults.
    """
    traits = obj.traits()
    for key, val in traits.items():
        if getattr(obj, key) != val.default_value:
            return False
    return True


def do_byteswap(arr):
    """
    Return a copy of an array, byteswapped.
    """
    return arr.byteswap().view(arr.dtype.newbyteorder('>'))


def ensure_bigendian(arr):
    """
    Ensure that a Numpy array is in big-endian order.

    Returns a copy if the endianness needed to be changed.
    """
    if needs_byteswap(arr):
        return do_byteswap(arr)
    return arr


if sys.byteorder == 'little':
    def needs_byteswap(arr):
        """
        Returns True if the array needs to be byteswapped.
        """
        order = arr.dtype.byteorder
        return order in ('<', '=')
else:
    def needs_byteswap(arr):
        """
        Returns True if the array needs to be byteswapped.
        """
        order = arr.dtype.byteorder
        return order == '<'


def ensure_native_endian(arr):
    """
    Ensure that a Numpy array is in native-endian order.

    Returns a copy if the endianness needed to be changed.
    """
    order = arr.dtype.byteorder

    if order != '=':
        return arr.byteswap().view(arr.dtype.newbyteorder('='))

    return arr


def unpack_bitflags(value, nbits):
    """
    Unpack a bitfield into its constituent parts.
    """
    return [bool(value & (1 << i)) for i in range(nbits)]


def pack_bitflags(*values):
    """
    Pack separate booleans back into a bit field.
    """
    result = 0
    for i, val in enumerate(values):
        if val:
            result |= (1 << i)
    return result



# -*- coding: utf-8 -*-


import io


import numpy as np
import pytest


import pytoshop
from pytoshop import image_data


def test_image_data_invalid_shape():
    with pytest.raises(ValueError):
        image_data.ImageData(channels=np.empty((52,), dtype=np.uint16))


def test_image_data_invalid_dtype():
    with pytest.raises(ValueError):
        image_data.ImageData(channels=np.empty((52,)))


def test_image_valid_size():
    pytoshop.PsdFile(
        version=1,
        num_channels=3,
        height=256,
        width=256,
        depth=8,
        color_mode=3,
        image_data=image_data.ImageData(
            channels=np.empty((3, 256, 256), np.uint8)))


def test_image_invalid_size():
    psd = pytoshop.PsdFile(
        version=1,
        num_channels=3,
        height=256,
        width=256,
        depth=8,
        color_mode=3,
        image_data=image_data.ImageData(
            channels=np.empty((3, 254, 256), np.uint8)))

    with pytest.raises(ValueError):
        psd.write(io.BytesIO())



# -*- coding: utf-8 -*-


"""
Coders and decoders for the various compression types in PSD.
"""


from __future__ import unicode_literals, absolute_import


import zlib


import numpy as np


from . import enums
from . import util


try:
    from . import packbits
except ImportError:
    pass


_decompress_params = """
    Parameters
    ----------
    data : bytes
        The raw bytes from the file.

    shape : 2-tuple of int
        The shape of the resulting array, ``(height, width)``.

    depth : enums.ColorDepth
        The bit depth of the image. See `enums.ColorDepth`.

    version : enums.Version
        The version of the PSD file. See `enums.Version`.

    Returns
    -------
    image : numpy array
        The image data as a Numpy array.  If *depth* is 1, the array
        is expanded to ``uint8`` with a byte per pixel.
"""


_compress_params = """
    Parameters
    ----------
    fd : file-like object
        Writable file-like object, open in binary mode.

    image : 2-D numpy array
        The image to compress.  Must be unsigned integer with 8, 16 or
        32 bits.  If *depth* is 1, the array should have dtype
        ``uint8`` with a byte per pixel.

    depth : enums.ColorDepth
        The bit depth of the image. See `enums.ColorDepth`.

    version : enums.Version
        The version of the PSD file. See `enums.Version`.
"""


_compress_constant_params = """
    Parameters
    ----------
    fd : file-like object
        Writable file-like object, open in binary mode.

    value : int
        The constant value in the generated virtual image.

    width : int
        The width of the image, in pixels.

    rows : int
        The number of rows in the image, in pixels.  This is
        ``height * num_channels``.

    depth : enums.ColorDepth
        The bit depth of the image. See `enums.ColorDepth`.

    version : enums.Version
        The version of the PSD file. See `enums.Version`.
"""


color_depth_dtype_map = {
    1: 'u1',
    8: 'u1',
    16: '>u2',
    32: '>u4'
}


color_depth_size_map = {
    1: 1,
    8: 1,
    16: 2,
    32: 4
}


def decompress_raw(data, shape, depth, version):
    """
    Converts raw data to a Numpy array.

{}
    """
    depth = enums.ColorDepth(depth)

    dtype = color_depth_dtype_map[depth]
    itemsize = color_depth_size_map[depth]

    # Truncate the data to a multiple of the dtype size
    data = data[:(len(data) // itemsize) * itemsize]

    arr = np.frombuffer(data, dtype)

    if depth == 1:
        # Unpack 1-bit image data
        arr = np.unpackbits(arr)

    # Make 2-dimensional
    image = arr.reshape(shape)
    return image


decompress_raw.__doc__ = decompress_raw.__doc__.format(_decompress_params)


def decompress_rle(data, shape, depth, version):
    """
    Decompress run length encoded data.

{}
    """
    output = packbits.decode(
        data, shape[0], shape[1], color_depth_size_map[depth], version)

    # Now pass along to the raw decoder to get a Numpy array
    return decompress_raw(output, shape, depth, version)


decompress_rle.__doc__ = decompress_rle.__doc__.format(_decompress_params)


def decompress_zip(data, shape, depth, version):
    """
    Decompress zip (zlib) encoded data.

{}
    """
    data = zlib.decompress(data)
    return decompress_raw(data, shape, depth, version)


decompress_zip.__doc__ = decompress_zip.__doc__.format(_decompress_params)


def decompress_zip_prediction(data, shape, depth, version):
    """
    Decompress zip (zlib) with prediction encoded data.

    Not supported for 1- or 32-bit images.

{}
    """
    if depth == 1:  # pragma: no cover
        raise ValueError(
            "zip with prediction is not supported for 1-bit images")
    elif depth == 32:
        raise ValueError(
            "zip with prediction is not implemented for 32-bit images")
    elif depth == 8:
        decoder = packbits.decode_prediction_8bit
    else:
        decoder = packbits.decode_prediction_16bit

    data = zlib.decompress(data)
    image = util.ensure_native_endian(
        decompress_raw(data, shape, depth, version))
    for i in range(len(image)):
        decoder(image[i].flatten())
    return image


decompress_zip_prediction.__doc__ = decompress_zip_prediction.__doc__.format(
    _decompress_params)


decompressors = {
    enums.Compression.raw: decompress_raw,
    enums.Compression.rle: decompress_rle,
    enums.Compression.zip: decompress_zip,
    enums.Compression.zip_prediction: decompress_zip_prediction
}


def decompress_image(data, compression, shape, depth, version):
    """
    Decompress data with the given compression.

    Parameters
    ----------
    data : bytes
        The raw bytes from the file.

    compression : enums.Compression
        The compression format to use.  See `enums.Compression`.

    shape : 2-tuple of int
        The shape of the resulting array, ``(height, width)``.

    depth : enums.ColorDepth
        The bit depth of the image. See `enums.ColorDepth`.

    version : enums.Version
        The version of the PSD file. See `enums.Version`.

    Returns
    -------
    image : numpy array
        The image data as a Numpy array.
    """
    compression = enums.Compression(compression)
    depth = enums.ColorDepth(depth)
    version = enums.Version(version)

    return decompressors[compression](data, shape, depth, version)


def normalize_image(image, depth):
    if depth == 1:
        image = np.packbits(image.flatten())
    return image


def compress_raw(fd, image, depth, version):
    """
    Write a Numpy array to raw bytes in a file.

{}
    """
    image = normalize_image(image, depth)
    if len(image.shape) == 2:
        if util.needs_byteswap(image):
            for row in image:
                row = util.do_byteswap(row)
                fd.write(row)
        else:
            fd.write(image)
    else:
        fd.write(image)


compress_raw.__doc__ = compress_raw.__doc__.format(_compress_params)


def compress_rle(fd, image, depth, version):
    """
    Write a Numpy array to a run length encoded stream.

{}
    """
    if depth == 1:  # pragma: no cover
        raise ValueError(
            "rle compression is not supported for 1-bit images")

    start = fd.tell()
    if version == 1:
        fd.seek(image.shape[0] * 2, 1)
        lengths = np.empty((len(image),), dtype='>u2')
    else:
        fd.seek(image.shape[0] * 4, 1)
        lengths = np.empty((len(image),), dtype='>u4')

    if util.needs_byteswap(image):
        for i, row in enumerate(image):
            row = util.do_byteswap(row)
            packed = packbits.encode(row)
            lengths[i] = len(packed)
            fd.write(packed)
    else:
        for i, row in enumerate(image):
            packed = packbits.encode(row)
            lengths[i] = len(packed)
            fd.write(packed)

    end = fd.tell()
    fd.seek(start)
    fd.write(lengths.tobytes())
    fd.seek(end)


compress_rle.__doc__ = compress_rle.__doc__.format(_compress_params)


def compress_zip(fd, image, depth, version):
    """
    Write a Numpy array to a zip (zlib) compressed stream.

{}
    """
    image = normalize_image(image, depth)
    if util.needs_byteswap(image):
        compressor = zlib.compressobj()
        for row in image:
            row = util.do_byteswap(row)
            fd.write(compressor.compress(row))
        fd.write(compressor.flush())
    else:
        fd.write(zlib.compress(image))


compress_zip.__doc__ = compress_zip.__doc__.format(_compress_params)


def compress_zip_prediction(fd, image, depth, version):
    """
    Write a Numpy array to a zip (zlib) with prediction compressed
    stream.

    Not supported for 1- or 32-bit images.

{}
    """
    if depth == 1:  # pragma: no cover
        raise ValueError(
            "zip with prediction is not supported for 1-bit images")
    elif depth == 32:  # pragma: no cover
        raise ValueError(
            "zip with prediction is not implemented for 32-bit images")
    elif depth == 8:
        encoder = packbits.encode_prediction_8bit
    elif depth == 16:
        encoder = packbits.encode_prediction_16bit

    compressor = zlib.compressobj()
    for row in image:
        encoder(row.flatten())
        row = util.ensure_bigendian(row)
        fd.write(compressor.compress(row))
    fd.write(compressor.flush())


compress_zip_prediction.__doc__ = compress_zip_prediction.__doc__.format(
    _compress_params)


compressors = {
    enums.Compression.raw: compress_raw,
    enums.Compression.rle: compress_rle,
    enums.Compression.zip: compress_zip,
    enums.Compression.zip_prediction: compress_zip_prediction
}


def compress_image(fd, image, compression, shape, num_channels, depth,
                   version):
    """
    Write an image with the given compression type.

    Parameters
    ----------
    fd : file-like object
        Writable file-like object, open in binary mode.

    image : 2-D numpy array or scalar
        The image to compress.  Must be unsigned integer with 8, 16 or
        32 bits.  If *depth* is 1, the array should have dtype
        ``uint8`` with a byte per pixel.  If a scalar, a "virtual"
        image will be used as if the image contained only that
        constant value.

    compression : enums.Compression
        The compression format to use.  See `enums.Compression`.

    shape : 2-tuple of int
        The shape of the image ``(height, width)``.  If *image* is an
        array, the *shape* is used to confirm it has the correct
        shape.  If *image* is a constant, *shape* is used to generate
        the virtual constant image.

    num_channels : int
        The number of color channels in the image.  If *image* is an
        array, the *num_channels* is used to confirm it has the
        correct number of channels.  If *image* is a constant,
        *num_channels* is used to generate the virtual constant image.

    depth : enums.ColorDepth
        The bit depth of the image. See `enums.ColorDepth`.  If
        *image* is an array, the *depth* is used to confirm it has the
        correct number of channels.  If *image* is a constant, *depth*
        is used to generate the virtual constant image.

    version : enums.Version
        The version of the PSD file. See `enums.Version`.
    """
    if isinstance(image, int):
        image = np.dtype(color_depth_dtype_map[depth]).type(image)

    dtype = image.dtype
    if dtype.kind != 'u':
        raise ValueError("Image array dtype must be unsigned int")
    if dtype.itemsize != color_depth_size_map[depth]:
        raise ValueError("Image array values of wrong size")

    if np.isscalar(image):
        width = shape[1]
        rows = shape[0] * num_channels
        return constant_compressors[compression](
            fd, image, width, rows, depth, version)
    else:
        acceptable_shapes = [
            (num_channels, shape[0], shape[1]),
            (num_channels * shape[0], shape[1])
        ]

        if image.shape not in acceptable_shapes:
            raise ValueError("Image is the wrong shape")

        image = np.asarray(image)
        image = util.ensure_native_endian(image)
        image = image.reshape((shape[0] * num_channels, shape[1]))
        return compressors[compression](fd, image, depth, version)


def _make_onebit_constant(value, width, rows):
    if value:
        value == 255
    else:
        value = 0
    return np.ones((width, rows), np.uint8) * value


def _make_constant_row(value, width, depth):
    return np.full((width,), value, dtype=color_depth_dtype_map[depth])


def compress_constant_raw(fd, value, width, rows, depth, version):
    """
    Write a virtual image containing a constant to a raw
    stream.

{}
    """
    if depth == 1:
        image = _make_onebit_constant(value, width, rows)
        compress_raw(fd, image, depth, version)
    else:
        row = _make_constant_row(value, width, depth)
        row = row.tobytes()
        for i in range(rows):
            fd.write(row)


compress_constant_raw.__doc__ = compress_constant_raw.__doc__.format(
    _compress_constant_params)


def compress_constant_rle(fd, value, width, rows, depth, version):
    """
    Write a virtual image containing a constant to a runlength-encoded
    stream.

{}
    """
    if depth == 1:  # pragma: no cover
        raise ValueError(
            "rle compression is not supported for 1-bit images")

    row = _make_constant_row(value, width, depth)
    packed = packbits.encode(row.tobytes())

    if version == 1:
        lengths = np.full((rows,), len(packed), dtype='>u2')
    else:
        lengths = np.full((rows,), len(packed), dtype='>u4')
    fd.write(lengths.tobytes())

    for i in range(rows):
        fd.write(packed)


compress_constant_rle.__doc__ = compress_constant_rle.__doc__.format(
    _compress_constant_params)


def compress_constant_zip(fd, value, width, rows, depth, version):
    """
    Write a virtual image containing a constant to a zip compressed
    stream.

{}
    """
    if depth == 1:
        image = _make_onebit_constant(value, width, rows)
        compress_zip(fd, image, depth, version)
    else:
        row = _make_constant_row(value, width, depth)
        row = row.tobytes()
        fd.write(zlib.compress(row * rows))


compress_constant_zip.__doc__ = compress_constant_zip.__doc__.format(
    _compress_constant_params)


def compress_constant_zip_prediction(
        fd, value, width, rows, depth, version):
    """
    Write a virtual image containing a constant to a zip with
    prediction compressed stream.

{}
    """
    if depth == 1:  # pragma: no cover
        raise ValueError(
            "zip with prediction is not supported for 1-bit images")
    elif depth == 32:  # pragma: no cover
        raise ValueError(
            "zip with prediction is not implemented for 32-bit images")
    elif depth == 8:
        encoder = packbits.encode_prediction_8bit
    elif depth == 16:
        encoder = packbits.encode_prediction_16bit

    row = _make_constant_row(value, width, depth)
    row = row.reshape((1, width))
    row = util.ensure_native_endian(row)
    encoder(row.flatten())
    row = util.ensure_bigendian(row)
    row = row.tobytes()
    fd.write(zlib.compress(row * rows))


compress_constant_zip_prediction.__doc__ = \
    compress_constant_zip_prediction.__doc__.format(
        _compress_constant_params)


constant_compressors = {
    enums.Compression.raw: compress_constant_raw,
    enums.Compression.rle: compress_constant_rle,
    enums.Compression.zip: compress_constant_zip,
    enums.Compression.zip_prediction: compress_constant_zip_prediction
}



# -*- coding: utf-8 -*-


import glob
import io
import os


import pytest


import pytoshop


path = os.path.join(os.path.dirname(__file__), 'psd_files', '*.psd')


@pytest.mark.parametrize("filename", glob.glob(path))
def test_files(filename):
    if filename.endswith('rt.psd'):
        return

    with open(filename, 'rb') as fd:
        f = pytoshop.read(fd)

        fd2 = io.BytesIO()
        f.write(fd2)

    fd2.seek(0)
    f = pytoshop.PsdFile.read(fd2)


@pytest.mark.parametrize("filename", glob.glob(path))
def test_convert_to_version2(filename):
    if filename.endswith('rt.psd'):
        return

    with open(filename, 'rb') as fd:
        f = pytoshop.read(fd)
        f.version = 2

        # Disable compression just to make these tests faster
        # for layer in f.layer_and_mask_info.layer_info.layer_records:
        #     for channel in layer.channels.values():
        #         channel.compression = 0

        fd2 = io.BytesIO()
        f.write(fd2)

    fd2.seek(0)
    f = pytoshop.PsdFile.read(fd2)



# -*- coding: utf-8 -*-


"""
Sections related to image layers.
"""


from __future__ import unicode_literals, absolute_import


import collections
import struct


import numpy as np
import traitlets as t


from . import blending_range
from . import codecs
from . import docs
from . import enums
from . import tagged_block
from . import util


class LayerMask(t.HasTraits):
    """
    Layer mask / adjustment layer data.
    """
    top = t.Int(
        help="Top of rectangle enclosing layer mask"
    )
    left = t.Int(
        help="Left of rectangle enclosing layer mask"
    )
    bottom = t.Int(
        help="Bottom of rectangle enclosing layer mask"
    )
    right = t.Int(
        help="Right of rectangle enclosing layer mask"
    )
    default_color = t.Bool(
        False,
        help="Default color for mask"
    )
    position_relative_to_layer = t.Bool(
        False,
        help="position relative to layer"
    )
    layer_mask_disabled = t.Bool(
        False,
        help="Layer mask disabled"
    )
    invert_layer_mask_when_blending = t.Bool(
        False,
        help="Invert layer mask when blending (obsolete)"
    )
    user_mask_from_rendering_other_data = t.Bool(
        False,
        help="Indicates that the user mask actually came from rendering other "
             "data"
    )
    user_mask_density = t.Int(
        None, min=0, max=255, allow_none=True,
        help="User mask density"
    )
    user_mask_feather = t.Float(
        None, allow_none=True,
        help="User mask feather"
    )
    vector_mask_density = t.Int(
        None, min=0, max=255, allow_none=True,
        help="Vector mask density"
    )
    vector_mask_feather = t.Float(
        None, allow_none=True,
        help="Vector mask feather"
    )
    real_flags = t.Int()
    real_background_color = t.Bool()
    real_top = t.Int()
    real_left = t.Int()
    real_bottom = t.Int()
    real_right = t.Int()

    @property
    def width(self):
        """
        Width of the mask layer.
        """
        return self.right - self.left

    @property
    def height(self):
        """
        Height of the mask layer.
        """
        return self.bottom - self.top

    @property
    def shape(self):
        """
        Shape of the mask layer ``(height, width)``.
        """
        return (self.height, self.width)

    @property
    def real_width(self):
        """
        Real width of the mask layer.
        """
        return self.real_right - self.real_left

    @property
    def real_height(self):
        """
        Real height of the mask layer.
        """
        return self.real_bottom - self.real_top

    @property
    def real_shape(self):
        """
        Real shape of the mask layer ``(height, width)``.
        """
        return (self.real_height, self.real_width)

    def length(self, header):
        if util.is_set_to_default(self):
            return 0
        length = 16 + 1 + 1
        mask_flags = self._get_mask_flags()
        if mask_flags:
            length += 1
            if self.user_mask_density is not None:
                length += 1
            if self.user_mask_feather is not None:
                length += 8
            if self.vector_mask_density is not None:
                length += 1
            if self.vector_mask_feather is not None:
                length += 8
        length += 1 + 1 + 16
        return length
    length.__doc__ = docs.length

    def total_length(self, header):
        return 4 + self.length(header)
    total_length.__doc__ = docs.total_length

    def _get_mask_flags(self):
        return util.pack_bitflags(
            self.user_mask_density is not None,
            self.user_mask_feather is not None,
            self.vector_mask_density is not None,
            self.vector_mask_feather is not None)

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        length = util.read_value(fd, 'I')
        d = {}
        end = fd.tell() + length
        util.log("length: {}, end: {}", length, end)

        if length == 0:
            return cls(**d)

        top, left, bottom, right = struct.unpack('>iiii', fd.read(16))
        d['top'] = top
        d['left'] = left
        d['bottom'] = bottom
        d['right'] = right

        util.log("position: ({}, {}, {}, {})", top, left, bottom, right)

        d['default_color'] = bool(util.read_value(fd, 'B'))

        flags = util.read_value(fd, 'B')
        (d['position_relative_to_layer'],
         d['layer_mask_disabled'],
         d['invert_layer_mask_when_blending'],
         d['user_mask_from_rendering_other_data']) = util.unpack_bitflags(
             flags, 4)

        util.log("default_color: {}, flags: {}", d['default_color'], flags)

        if length == 20:
            util.log("done early")
            fd.seek(end)
            return cls(**d)

        if flags & 16:
            mask_parameters = util.read_value(fd, 'B')
            (has_user_mask_density,
             has_user_mask_feather,
             has_vector_mask_density,
             has_vector_mask_feather) = util.unpack_bitflags(
                 mask_parameters, 4)
            if has_user_mask_density:
                d['user_mask_density'] = util.read_value(fd, 'B')
            if has_user_mask_feather:
                d['user_mask_feather'] = util.read_value(fd, 'd')
            if has_vector_mask_density:
                d['vector_mask_density'] = util.read_value(fd, 'B')
            if has_vector_mask_feather:
                d['vector_mask_feather'] = util.read_value(fd, 'd')

        d['real_flags'] = util.read_value(fd, 'B')
        d['real_user_mask_background'] = bool(util.read_value(fd, 'B'))

        util.log(
            "real_flags: {}, real_user_mask_background: {}",
            d['real_flags'], d['real_user_mask_background']
        )

        top, left, bottom, right = struct.unpack('>iiii', fd.read(16))
        d['real_top'] = top
        d['real_left'] = left
        d['real_bottom'] = bottom
        d['real_right'] = right

        util.log(
            "real position: ({}, {}, {}, {})",
            top, left, bottom, right
        )

        fd.seek(end)

        return cls(**d)
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        def write_rectangle(top, left, bottom, right):
            fd.write(
                struct.pack(
                    '>iiii', top, left, bottom, right))

        def write_default_color(color):
            if color:
                util.write_value(fd, 'B', 255)
            else:
                util.write_value(fd, 'B', 0)

        util.write_value(fd, 'I', self.length(header))
        if util.is_set_to_default(self):
            return

        write_rectangle(self.top, self.left, self.bottom, self.right)
        write_default_color(self.default_color)

        mask_flags = self._get_mask_flags()

        flags = util.pack_bitflags(
            self.position_relative_to_layer,
            self.layer_mask_disabled,
            self.invert_layer_mask_when_blending,
            self.user_mask_from_rendering_other_data,
            mask_flags)

        util.write_value(fd, 'B', flags)

        if mask_flags:
            util.write_value(fd, 'B', mask_flags)

            if self.user_mask_density is not None:
                util.write_value(fd, 'B', self.user_mask_density)
            if self.user_mask_feather is not None:
                util.write_value(fd, 'd', self.user_make_feather)
            if self.vector_mask_density is not None:
                util.write_value(fd, 'B', self.vector_mask_density)
            if self.vector_mask_feather is not None:
                util.write_value(fd, 'd', self.vector_mask_feather)

        util.write_value(fd, 'B', self.real_flags)
        write_default_color(self.real_background_color)
        write_rectangle(self.real_top, self.real_left,
                        self.real_bottom, self.real_right)
    write.__doc__ = docs.write


class ChannelImageData(t.HasTraits):
    """
    A single plane of channel image data.
    """
    compression = t.Enum(
        list(enums.Compression),
        default_value=enums.Compression.rle,
        help="Compression method. See `enums.Compression`."
    )

    def __init__(self, image=None, fd=None, offset=None, size=None,
                 shape=None, depth=None, version=None,
                 compression=enums.Compression.raw):
        t.HasTraits.__init__(self, compression=compression)
        if image is not None:
            if (fd is not None or offset is not None or size is not None or
                    shape is not None or depth is not None or
                    version is not None):
                raise ValueError(
                    "May not provide both image and other parameters")
            self._image = image
        else:
            if image is not None:
                raise ValueError(
                    "May not provide both image and other parameters")
            self._image = None
            self._fd = fd
            self._offset = offset
            self._size = size
            self._shape = shape
            self._depth = depth
            self._version = version

    @property
    def image(self):
        if self._image is not None:
            return self._image
        tell = self._fd.tell()
        try:
            self._fd.seek(self._offset)
            data = self._fd.read(self._size)
            return codecs.decompress_image(
                data, self.compression, self._shape, self._depth,
                self._version)
        finally:
            self._fd.seek(tell)

    @image.setter
    def image(self, image):
        self._image = image

    @property
    def shape(self):
        if self._image is not None:
            return self._image.shape
        return self._shape

    @property
    def dtype(self):
        return np.dtype(codecs.color_depth_dtype_map[self._depth])

    @classmethod
    @util.trace_read
    def read(cls, fd, header, shape, size):
        compression = util.read_value(fd, 'H')
        util.log("compression: {}", enums.Compression(compression))
        offset = fd.tell()
        fd.seek(size, 1)

        return cls(fd=fd, offset=offset, size=size, shape=shape,
                   depth=header.depth, version=header.version,
                   compression=compression)
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header, shape):
        start = fd.tell()
        util.write_value(fd, 'H', self.compression)
        if self._image is not None:
            codecs.compress_image(
                fd, self.image, self.compression, shape, 1,
                header.depth, header.version)
        else:
            tell = self._fd.tell()
            try:
                self._fd.seek(self._offset)
                data = self._fd.read(self._size)
            finally:
                self._fd.seek(tell)
            fd.write(data)
        return fd.tell() - start
    write.__doc__ = docs.write


class LayerRecord(t.HasTraits):
    """
    Layer record.

    There is one of these per logical layer in the file.
    """
    top = t.Int(
        help="Top of the rectangle containing the layer."
    )
    left = t.Int(
        help="Left of the rectangle containing the layer."
    )
    bottom = t.Int(
        help="Bottom of the rectangle containing the layer."
    )
    right = t.Int(
        help="Right of the rectangle containing the layer."
    )
    blend_mode = t.Enum(
        list(enums.BlendMode), default_value=enums.BlendMode.normal,
        help="Blend mode. See `enums.BlendMode`"
    )
    opacity = t.Int(
        255, min=0, max=255,
        help="Opacity. 0=transparent, 255=opaque"
    )
    clipping = t.Bool(
        False,
        help="Clipping. False=base, True=non-base"
    )
    transparency_protected = t.Bool(
        False,
        help="Transparency protected"
    )
    visible = t.Bool(
        True,
        help="Visible"
    )
    pixel_data_irrelevant = t.Bool(
        False,
        help="Pixel data is irrelevant to appearance of document"
    )
    mask = t.Instance(
        LayerMask,
        help="`LayerMask`"
    )
    blending_ranges = t.Instance(
        blending_range.BlendingRanges,
        help='`blending_range.BlendingRanges`'
    )
    name = t.Unicode(
        min=0, max=255, default_value='',
        help="Name of layer"
    )
    channels = t.Dict(
        help="Dictionary from `enums.ChannelId` to `ChannelImageData`."
    )
    blocks = t.List(
        t.Instance(tagged_block.TaggedBlock),
        help="List of `tagged_block.TaggedBlock` items with additional "
             "information about this layer."
    )

    @t.default('mask')
    def _default_mask(self):
        return LayerMask()

    @t.default('blending_ranges')
    def _default_blending_ranges(self):
        return blending_range.BlendingRanges()

    @t.validate('channels')
    def _validate_channels(self, proposal):
        value = proposal['value']

        for key, val in value.items():
            try:
                enums.ChannelId(key)
            except ValueError as e:
                raise t.TraitError(str(e))

            if not isinstance(val, ChannelImageData):
                raise t.TraitError(
                    "Each channel must be ChannelImageData instance")

        value = collections.OrderedDict(
            sorted([(k, v) for (k, v) in value.items()]))

        return value

    @property
    def width(self):
        """
        Width of the layer.
        """
        return self.right - self.left

    @property
    def height(self):
        """
        Height of the layer.
        """
        return self.bottom - self.top

    @property
    def shape(self):
        """
        Shape of the layer ``(height, width)``.
        """
        return (self.height, self.width)

    @property
    def blocks_map(self):
        """
        A mapping from tagged block codes to
        `tagged_block.TaggedBlock` instances.

        This is a convenience to more easily get associated tagged
        blocks.
        """
        return dict((x.code, x) for x in self.blocks)

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        top, left, bottom, right = struct.unpack('>iiii', fd.read(16))

        util.log("position: ({}, {}, {}, {})", top, left, bottom, right)

        num_channels = util.read_value(fd, 'H')
        channel_ids = []
        channel_data_lengths = []
        for i in range(num_channels):
            channel_ids.append(util.read_value(fd, 'h'))
            if header.version == 1:
                channel_data_lengths.append(util.read_value(fd, 'I'))
            else:
                channel_data_lengths.append(util.read_value(fd, 'Q'))

        util.log(
            "num_channels: {}, channel_ids: {}, channel_data_lengths: {}",
            num_channels, channel_ids, channel_data_lengths
        )

        blend_mode_signature = fd.read(4)
        if blend_mode_signature != b'8BIM':
            raise ValueError(
                "Invalid blend mode signature '{}'".format(
                    blend_mode_signature))

        blend_mode = fd.read(4)
        opacity = util.read_value(fd, 'B')
        clipping = bool(util.read_value(fd, 'B'))
        flags = util.read_value(fd, 'B')
        (transparency_protected,
         visible,
         _,
         _,
         pixel_data_irrelevant) = util.unpack_bitflags(flags, 5)
        visible = not visible
        fd.seek(1, 1)  # filler

        util.log(
            "blend_mode: {}, opacity: {}, clipping: {}, flags: {}",
            blend_mode, opacity, clipping, flags
        )

        extra_length = util.read_value(fd, 'I')
        end = fd.tell() + extra_length

        util.log("extra_length: {}, end: {}", extra_length, end)

        mask = LayerMask.read(fd, header)
        blending_ranges = blending_range.BlendingRanges.read(
            fd, header, num_channels)
        name = util.read_pascal_string(fd, 4)

        util.log("name: '{}'", name)

        blocks = []
        while fd.tell() < end:
            blocks.append(
                tagged_block.TaggedBlock.read(fd, header))
        fd.seek(end)

        result = cls(
            header,
            top=top,
            left=left,
            bottom=bottom,
            right=right,
            blend_mode=blend_mode,
            opacity=opacity,
            clipping=clipping,
            transparency_protected=transparency_protected,
            visible=visible,
            pixel_data_irrelevant=pixel_data_irrelevant,
            mask=mask,
            blending_ranges=blending_ranges,
            name=name,
            blocks=blocks)

        result._channel_data_lengths = channel_data_lengths
        result._channel_ids = channel_ids
        return result
    read.__func__.__doc__ = docs.read

    def read_channel_data(self, fd, header):
        """
        Read the `ChannelImageData` for this layer.
        """
        channels = collections.OrderedDict()
        for channel_id, channel_length in zip(
                self._channel_ids, self._channel_data_lengths):
            if channel_id == enums.ChannelId.user_layer_mask:
                shape = self.mask.shape
            elif channel_id == enums.ChannelId.real_user_layer_mask:
                shape = self.mask.real_shape
            else:
                shape = self.shape
            channels[channel_id] = ChannelImageData.read(
                fd, header, shape, channel_length - 2)
        self.channels = channels

    @util.trace_write
    def write(self, fd, header):
        fd.write(struct.pack('>iiii',
                 self.top, self.left, self.bottom, self.right))
        util.write_value(fd, 'H', len(self.channels))
        self.channel_lengths_offset = fd.tell()
        if header.version == 1:
            fd.seek(6 * len(self.channels), 1)
        else:
            fd.seek(10 * len(self.channels), 1)
        fd.write(b'8BIM')
        fd.write(self.blend_mode)
        util.write_value(fd, 'B', self.opacity)
        util.write_value(fd, 'B', int(self.clipping))
        flags = util.pack_bitflags(
            self.transparency_protected,
            not self.visible,
            False,
            True,
            self.pixel_data_irrelevant)
        util.write_value(fd, 'B', flags)
        fd.write(b'\0')  # filler

        extra_length = (
            self.mask.total_length(header) +
            self.blending_ranges.total_length(header) +
            util.pascal_string_length(self.name, 4) +
            sum(x.total_length(header) for x in self.blocks)
        )
        util.write_value(fd, 'I', extra_length)
        self.mask.write(fd, header)
        self.blending_ranges.write(fd, header)
        util.write_pascal_string(fd, self.name, 4)
        for block in self.blocks:
            block.write(fd, header)
    write.__doc__ = docs.write

    def write_channel_data(self, fd, header):
        """
        Write the `ChannelImageData` for this layer.
        """
        lengths = []
        for channel_id, data in self.channels.items():
            if channel_id == enums.ChannelId.user_layer_mask:
                shape = self.mask.shape
            elif channel_id == enums.ChannelId.real_user_layer_mask:
                shape = self.mask.real_shape
            else:
                shape = self.shape
            lengths.append(data.write(fd, header, shape))

        offset = fd.tell()
        fd.seek(self.channel_lengths_offset)
        for channel_id, length in zip(self.channels.keys(), lengths):
            util.write_value(fd, 'h', channel_id)
            if header.version == 1:
                util.write_value(fd, 'I', length)
            else:
                util.write_value(fd, 'Q', length)
        fd.seek(offset)


class LayerInfo(t.HasTraits):
    """
    A set of `LayerRecord` instances.
    """
    layer_records = t.List(
        t.Instance(LayerRecord),
        help="List of `LayerRecord` instances"
    )
    use_alpha_channel = t.Bool(
        False,
        help="Indicates that the first channel contains transparency data for "
             "the merged result."
    )

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        if header.version == 1:
            length = util.read_value(fd, 'I')
        else:
            length = util.read_value(fd, 'Q')
        end = fd.tell() + length

        util.log("length: {}, end: {}", length, end)

        if length > 0:
            layer_count = util.read_value(fd, 'h')
            if layer_count < 0:
                layer_count = abs(layer_count)
                use_alpha_channel = True
            else:
                use_alpha_channel = False

            util.log("layer_count: {}, use_alpha_channel: {}",
                     layer_count, use_alpha_channel)

            layer_records = [
                LayerRecord.read(fd, header) for i in range(layer_count)
            ]
            for layer in layer_records:
                layer.read_channel_data(fd, header)

            fd.seek(end)

            return cls(
                layer_records=layer_records,
                use_alpha_channel=use_alpha_channel)
        else:
            return cls()
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        start = fd.tell()
        if header.version == 1:
            fd.seek(4, 1)
        else:
            fd.seek(8, 1)

        layer_count = len(self.layer_records)
        if layer_count == 0:
            return
        if self.use_alpha_channel:
            layer_count *= -1
        util.write_value(fd, 'h', layer_count)
        for layer in self.layer_records:
            layer.write(fd, header)
        for layer in self.layer_records:
            layer.write_channel_data(fd, header)

        end = fd.tell()
        fd.seek(start)
        if header.version == 1:
            util.write_value(fd, 'I', end - start - 4)
        else:
            util.write_value(fd, 'Q', end - start - 8)
        fd.seek(end)
    write.__doc__ = docs.write


class GlobalLayerMaskInfo(t.HasTraits):
    """
    Global layer mask info.
    """
    overlay_color_space = t.Bytes(
        b'\0' * 10, min=10, max=10,
        help="Undocumented"
    )
    opacity = t.Int(
        100, min=0, max=100,
        help="Opacity. 0=transparent, 100=opaque"
    )
    kind = t.Enum(
        list(enums.LayerMaskKind),
        default_value=enums.LayerMaskKind.use_value_stored_per_layer,
        help="Layer mask kind. See `enums.LayerMaskKind`"
    )

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        length = util.read_value(fd, 'I')
        end = fd.tell() + length
        util.log("length: {}, end: {}", length, end)
        if length == 0:
            return cls()

        overlay_color_space = fd.read(10)
        opacity = util.read_value(fd, 'H')
        kind = util.read_value(fd, 'B')

        util.log(
            "overlay_color_space: {}, opacity: {}, kind: {}",
            overlay_color_space, opacity, kind
        )

        fd.seek(end)

        return cls(
            overlay_color_space=overlay_color_space,
            opacity=opacity,
            kind=kind)
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        if util.is_set_to_default(self):
            util.write_value(fd, 'I', 0)
        else:
            util.write_value(fd, 'I', 16)
            fd.write(self.overlay_color_space)
            util.write_value(fd, 'H', self.opacity)
            util.write_value(fd, 'B', self.kind)
            fd.write(b'\0\0\0')  # filler
    write.__doc__ = docs.write


class LayerAndMaskInfo(t.HasTraits):
    """
    Layer and mask information section.
    """
    layer_info = t.Instance(
        LayerInfo,
        help="Layer info. See `LayerInfo`."
    )
    global_layer_mask_info = t.Instance(
        GlobalLayerMaskInfo, allow_none=True,
        help="Global layer mask info. See `GlobalLayerMaskInfo`."
    )
    additional_layer_info = t.List(
        t.Instance(tagged_block.TaggedBlock),
        help="List of additional layer info. See `TaggedBlock`."
    )

    @t.default('layer_info')
    def _default_layer_info(self):
        return LayerInfo()

    @property
    def additional_layer_info_map(self):
        """
        A mapping from tagged block codes to
        `tagged_block.TaggedBlock` instances.

        This is a convenience to more easily get associated tagged
        blocks.
        """
        return dict((x.code, x) for x in self.additional_layer_info)

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        if header.version == 1:
            length = util.read_value(fd, 'I')
        else:
            length = util.read_value(fd, 'Q')
        end = fd.tell() + length

        util.log("length: {}, end: {}", length, end)

        layer_info = LayerInfo.read(fd, header)

        global_layer_mask_info = None
        additional_layer_info = []
        if fd.tell() < end:
            global_layer_mask_info = GlobalLayerMaskInfo.read(fd, header)

            while fd.tell() < end:
                additional_layer_info.append(
                    tagged_block.TaggedBlock.read(fd, header, 4))

        return cls(layer_info=layer_info,
                   global_layer_mask_info=global_layer_mask_info,
                   additional_layer_info=additional_layer_info)
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        start = fd.tell()

        if header.version == 1:
            fd.seek(4, 1)
        else:
            fd.seek(8, 1)

        self.layer_info.write(fd, header)
        if (self.global_layer_mask_info is not None or
                len(self.additional_layer_info)):
            if self.global_layer_mask_info is None:
                global_layer_mask_info = GlobalLayerMaskInfo()
            else:
                global_layer_mask_info = self.global_layer_mask_info
            global_layer_mask_info.write(fd, header)
            for layer_info in self.additional_layer_info:
                layer_info.write(fd, header, 4)

        end = fd.tell()
        fd.seek(start)
        if header.version == 1:
            util.write_value(fd, 'I', end - start - 4)
        else:
            util.write_value(fd, 'Q', end - start - 8)
        fd.seek(end)
    write.__doc__ = docs.write



# -*- coding: utf-8 -*-


"""
Convert a PSD file to/from nested layers.
"""


from collections import OrderedDict
import sys


import numpy as np
import traitlets as t


from .. import core
from .. import enums
from .. import image_resources
from .. import layers as l
from .. import path
from .. import tagged_block


class Layer(t.HasTraits):
    """
    Base class of all layers.
    """
    name = t.Unicode(
        help="The name of the layer"
    )
    visible = t.Bool(
        True,
        help="Is layer visible?"
    )
    opacity = t.Int(
        255, min=0, max=255,
        help="Opacity. 0=transparent, 255=opaque"
    )
    group_id = t.Int(
        0, min=0, max=((1 << 16) - 1),
        help="Linked layer id"
    )


class Group(Layer):
    """
    A `Layer` that may contain other `Layer` instances.
    """
    blend_mode = t.Enum(
        list(enums.BlendMode),
        default_value=enums.BlendMode.pass_through,
        help="blend mode"
    )
    layers = t.List(
        t.Instance(Layer),
        help="List of sublayers"
    )
    closed = t.Bool(
        True,
        help="Is layer closed in GUI?"
    )


class Image(Layer):
    """
    A `Layer` containing image data, i.e. a leaf node.
    """
    blend_mode = t.Enum(
        list(enums.BlendMode),
        default_value=enums.BlendMode.normal,
        help="blend mode"
    )
    top = t.Int(
        0,
        help="The top of the layer, in pixels."
    )
    left = t.Int(
        0,
        help="The left of the layer, in pixels."
    )
    bottom = t.Int(
        None, allow_none=True,
        help="The bottom of the layer, in pixels. If not provided, "
             "will be automatically determined from channel data."
    )
    right = t.Int(
        None, allow_none=True,
        help="The right of the layer, in pixels. If not provided, "
             "will be automatically determined from channel data."
    )
    channels = t.TraitType(
        help="""
        The channel image data. May be one of the following:
        - A dictionary from `enums.ChannelId` to 2-D numpy arrays.
        - A 3-D numpy array of the shape (num_channels, height, width)
        - A list of numpy arrays where each is a channel.
        """
    )

    @t.validate('channels')
    def _validate_channels(self, proposal):
        value = proposal['value']
        if isinstance(value, dict):
            for key in value.keys():
                enums.ChannelId(key)
            return value

        if isinstance(value, np.ndarray):
            if len(value.shape) == 3:
                return dict((i, plane) for (i, plane) in enumerate(value))
            else:
                return {0: value}

        if isinstance(value, list):
            return dict((i, plane) for (i, plane) in enumerate(value))

        return {0: value}


def _iterate_all_images(layers):
    """
    Iterate over all `Image` instances in a hierarchy of `Layer`
    instances.
    """
    if isinstance(layers, list):
        for layer in layers:
            for sublayer in _iterate_all_images(layer):
                yield sublayer
    if isinstance(layers, Group):
        for sublayer in _iterate_all_images(layers.layers):
            yield sublayer
    elif isinstance(layers, Image):
        yield layers


def pprint_layers(layers, indent=0):
    """
    Pretty-print a hierarchy of `Layer` instances.
    """
    for layer in layers:
        if isinstance(layer, Group):
            print(('  ' * indent) + '< {}'.format(layer.name))
            pprint_layers(layer.layers, indent+1)
            print(('  ' * indent) + '>')
        elif isinstance(layer, Image):
            print(('  ' * indent) + '< {} ({}, {}, {}, {}) >'.format(
                layer.name, layer.top, layer.left, layer.bottom, layer.right))


def psd_to_nested_layers(psdfile):
    """
    Convert a `PsdFile` instance to a hierarchy of nested `Layer`
    instances.

    Parameters
    ----------
    psdfile : PsdFile
        A parsed PSD file.

    Returns
    -------
    layers : list of `Layer` instances
        A representation of the parsed hierarchy from the file.
    """
    if not isinstance(psdfile, core.PsdFile):
        raise TypeError("psdfile must be a pytoshop.core.PsdFile instance")

    group_ids_block = psdfile.image_resources.get_block(
        enums.ImageResourceID.layers_group_info)
    if group_ids_block is not None:
        group_ids = group_ids_block.group_ids

    layers = psdfile.layer_and_mask_info.layer_info.layer_records

    root = Group()
    group_stack = [root]

    for index, layer in reversed(list(enumerate(layers))):
        current_group = group_stack[-1]

        blocks = layer.blocks_map

        name = blocks.get(b'luni', layer).name
        divider = blocks.get(b'lsct', blocks.get(b'lsdk'))
        visible = layer.visible
        opacity = layer.opacity
        blend_mode = layer.blend_mode
        if group_ids is not None:
            group_id = int(group_ids[index])
        else:
            group_id = None

        if divider is not None:
            if divider.type in (enums.SectionDividerSetting.closed,
                                enums.SectionDividerSetting.open):
                # group begins
                group = Group(
                    name=name,
                    closed=(
                        divider.type == enums.SectionDividerSetting.closed),
                    blend_mode=blend_mode,
                    visible=visible,
                    opacity=opacity,
                    group_id=group_id
                )
                group_stack.append(group)
                current_group.layers.append(group)

            elif divider.type == enums.SectionDividerSetting.bounding:
                # group ends
                if len(group_stack) == 1:
                    layers = group_stack[0].layers
                    layer = layers[0]

                    group = Group(
                        name=layer.name,
                        closed=False,
                        blend_mode=layer.blend_mode,
                        visible=layer.visible,
                        opacity=layer.opacity,
                        layers=layers[1:],
                        group_id=group_id
                    )

                    group_stack[0].layers = [group]

                else:
                    finished_group = group_stack.pop()
                    assert finished_group is not root

            else:
                raise ValueError("Invalid state")

        else:
            layer = Image(
                name=name,
                top=layer.top,
                left=layer.left,
                bottom=layer.bottom,
                right=layer.right,
                channels=layer.channels,
                blend_mode=blend_mode,
                visible=visible,
                opacity=opacity,
                group_id=group_id
            )
            current_group.layers.append(layer)

    return root.layers


def _flatten_layers(layers, flat_layers, group_ids, compression, vector_mask):
    for layer in layers:
        if isinstance(layer, Group):
            if layer.closed:
                divider_type = enums.SectionDividerSetting.closed
            else:
                divider_type = enums.SectionDividerSetting.open

            name_source = len(flat_layers)

            flat_layers.append(
                l.LayerRecord(
                    name=layer.name,
                    blend_mode=layer.blend_mode,
                    opacity=layer.opacity,
                    visible=layer.visible,
                    blocks=[
                        tagged_block.UnicodeLayerName(name=layer.name),
                        tagged_block.SectionDividerSetting(type=divider_type),
                        tagged_block.LayerId(id=len(flat_layers))
                    ]
                )
            )

            group_ids.append(layer.group_id)

            _flatten_layers(
                layer.layers, flat_layers, group_ids, compression, vector_mask)

            flat_layers.append(
                l.LayerRecord(
                    blocks=[
                        tagged_block.SectionDividerSetting(
                            type=enums.SectionDividerSetting.bounding),
                        tagged_block.LayerNameSource(id=name_source)
                    ]
                )
            )

            group_ids.append(0)

        elif isinstance(layer, Image):
            channels = OrderedDict()
            for id, im in layer.channels.items():
                if isinstance(im, l.ChannelImageData):
                    channels[id] = im
                else:
                    channels[id] = l.ChannelImageData(
                        image=im, compression=compression)
            if (enums.ChannelId.transparency in channels and
                    np.all(channels[enums.ChannelId.transparency].image == 0)):
                continue

            blocks = [
                tagged_block.UnicodeLayerName(name=layer.name),
                tagged_block.LayerId(id=len(flat_layers))
            ]

            if vector_mask:
                blocks.append(
                    tagged_block.VectorMask(
                        path_resource=path.PathResource.from_rect(
                            layer.top + 5, layer.left + 5,
                            layer.bottom - 5, layer.right - 5
                        )
                    )
                )
            else:
                if enums.ChannelId.transparency not in channels:
                    channels[enums.ChannelId.transparency] = \
                        l.ChannelImageData(
                            image=-1, compression=compression)

            flat_layers.append(
                l.LayerRecord(
                    top=layer.top,
                    left=layer.left,
                    bottom=layer.bottom,
                    right=layer.right,
                    name=layer.name,
                    blend_mode=layer.blend_mode,
                    opacity=layer.opacity,
                    visible=layer.visible,
                    channels=channels,
                    blocks=blocks
                )
            )

            group_ids.append(layer.group_id)

    return flat_layers, group_ids


def _adjust_positions(layers):
    top = sys.maxsize
    left = sys.maxsize
    bottom = -sys.maxsize
    right = -sys.maxsize
    for image in _iterate_all_images(layers):
        top = min(image.top, top)
        left = min(image.left, left)
        bottom = max(image.bottom, bottom)
        right = max(image.right, right)

    yoffset = -top
    xoffset = -left

    for image in _iterate_all_images(layers):
        image.top += yoffset
        image.left += xoffset
        image.bottom += yoffset
        image.right += xoffset

    width = right - left
    height = bottom - top

    return width, height


def _determine_channels_and_depth(layers, depth):
    num_channels = 0
    for image in _iterate_all_images(layers):
        for index, channel in image.channels.items():
            if np.isscalar(channel):
                continue
            num_channels = max(num_channels, index + 1)
            channel_depth = channel.dtype.itemsize * 8
            if depth is None:
                depth = channel_depth
            elif depth != channel_depth:
                raise ValueError("Different image depths in input")

    if num_channels == 0 or depth is None:
        raise ValueError("Can't determine num channels or depth")

    return num_channels, depth


def _update_sizes(layers):
    for image in _iterate_all_images(layers):
        if len(image.channels) == 0:
            width = 0
            height = 0
        else:
            if image.bottom is not None and image.top is not None:
                height = image.bottom - image.top
            else:
                height = None
            if image.right is not None and image.left is not None:
                width = image.right - image.left
            else:
                width = None
            for index, channel in image.channels.items():
                if np.isscalar(channel):
                    continue
                if height is None:
                    height = channel.shape[0]
                if width is None:
                    width = channel.shape[1]
                if (height, width) != channel.shape:
                    raise ValueError(
                        "Channels in image have different shapes or do not "
                        "match layer"
                    )
            if height is None or width is None:
                raise ValueError(
                    "Can't determine shape.  Set it explicitly."
                )

        if image.bottom is None:
            image.bottom = (image.top or 0) + height

        if image.right is None:
            image.right = (image.left or 0) + width


def nested_layers_to_psd(
        layers,
        color_mode,
        version=enums.Version.version_1,
        compression=enums.Compression.rle,
        depth=None,
        size=None,
        vector_mask=False):
    """
    Convert a hierarchy of nested `Layer` instances to a `PsdFile`.

    Parameters
    ----------
    layers : list of `Layer` instances
        The hierarchy of layers we want to create.

    color_mode : `enums.ColorMode`
        The color mode of the resulting PSD file (as well as the input
        image data).

    version : `enums.Version`, optional
        The version of the PSD spec to follow.

    compression : `enums.Compression`, optional
        The method of image compression to use for the layer image
        data.

    depth : `enums.ColorDepth`, optional
        The color depth of the resulting image.  Must match the color
        depth of the data passed in.  If not provided, the color depth
        will be automatically determined from the passed-in data.

    size : 2-tuple of int, optional
        The shape in the form ``(height, width)`` of the resulting PSD
        file.  If not provided, the height and width will be set to
        include all passed in layers, and the layers themselves will
        be adjusted so that none fall outside of the image.

    vector_mask : bool, optional
        When `True`, the mask for the layer will be a vector
        rectangle.  This results in much smaller file sizes, but is
        not quite as accurately rendered by Photoshop.  When `False`,
        a raster mask is used.

    Returns
    -------
    psdfile : PsdFile
        The resulting PSD file.
    """
    try:
        next(_iterate_all_images(layers))
    except StopIteration:
        raise ValueError("No images found in layers")

    _update_sizes(layers)

    num_channels, depth = _determine_channels_and_depth(layers, depth)

    if size is None:
        width, height = _adjust_positions(layers)
    else:
        width, height = size

    flat_layers, group_ids = _flatten_layers(
        layers, [], [], compression, vector_mask)

    flat_layers = flat_layers[::-1]
    group_ids = group_ids[::-1]

    f = core.PsdFile(
        version=version,
        num_channels=num_channels,
        height=height,
        width=width,
        depth=depth,
        color_mode=color_mode,
        layer_and_mask_info=l.LayerAndMaskInfo(
            layer_info=l.LayerInfo(
                layer_records=flat_layers
            )
        ),
        image_resources=image_resources.ImageResources(
            blocks=[image_resources.LayersGroupInfo(
                group_ids=group_ids)]
        )
    )

    return f



# -*- coding: utf-8 -*-


"""
The core objects, including the `PsdFile` and its `Header`.
"""


from __future__ import unicode_literals, absolute_import


import struct


import traitlets as t


from . import color_mode
from . import docs
from . import enums
from . import image_data
from . import image_resources
from . import layers
from . import util


class Header(t.HasTraits):
    """
    Manages the header at the start of a PSD/PSB file.
    """

    version = t.Enum(
        list(enums.Version),
        default_value=enums.Version.version_1,
        help="The version of the file format. See `enums.Version`."
    )
    num_channels = t.Int(
        1, min=1, max=56,
        help="Number of color channels in the file."
    )
    height = t.Int(
        1, min=1, max=300000,
        help="Height of the image (in pixels)."
    )
    width = t.Int(
        1, min=1, max=300000,
        help="Width of the image (in pixels)."
    )
    depth = t.Enum(
        list(enums.ColorDepth),
        default_value=enums.ColorDepth.depth8,
        help="Number of bits per channel. See `enums.ColorDepth`."
    )
    color_mode = t.Enum(
        list(enums.ColorMode),
        default_value=enums.ColorMode.rgb,
        help="Color mode of the file. See `enums.ColorMode`."
    )

    max_size_mapping = {
        1: 30000,
        2: 300000
    }

    @t.observe('width', 'height')
    def _check_size(self, change):
        if change['new'] > self.max_size_mapping[self.version]:
            raise t.TraitError(
                '{} is too large for version {}. Must be <= {}'.format(
                    change['name'],
                    self.version,
                    self.max_size_mapping[self.version]))

    @property
    def shape(self):
        return (self.height, self.width)

    @classmethod
    @util.trace_read
    def header_read(cls, fd):
        data = fd.read(26)

        (signature, version, _reserved, num_channels,
         height, width, depth, color_mode) = struct.unpack('>4sH6sHIIHH', data)

        if signature != b'8BPS':
            raise ValueError("Invalid signature '{}'".format(signature))

        util.log(
            'version: {}, num_channels: {}, '
            'width: {}, height: {}, depth: {}, '
            'color_mode: {}',
            enums.Version(version), num_channels, width, height,
            depth, enums.ColorMode(color_mode)
        )

        return cls(version=version, num_channels=num_channels,
                   width=width, height=height, depth=depth,
                   color_mode=color_mode)
    header_read.__func__.__doc__ = docs.read_single

    read = header_read

    @util.trace_write
    def write(self, fd):
        """
        Write to a file-like object.

        Parameters
        ----------
        fd : file-like object
            Must be writable, seekable and open in binary mode.
        """
        data = struct.pack(
            '>4sH6sHIIHH', b'8BPS', self.version, b'',
            self.num_channels, self.height, self.width, self.depth,
            self.color_mode)
        fd.write(data)
    write.__doc__ = docs.write_single


class PsdFile(Header):
    """
    Represents an entire PSD file.
    """

    color_mode_data = t.Instance(
        color_mode.ColorModeData,
        help='Color mode data section. See `color_mode.ColorModeData`.'
    )
    image_resources = t.Instance(
        image_resources.ImageResources,
        help='Image resources. See `image_resources.ImageResources`.'
    )
    layer_and_mask_info = t.Instance(
        layers.LayerAndMaskInfo,
        help='Layer and mask info. See `layers.LayerAndMaskInfo`.'
    )
    image_data = t.Instance(
        image_data.ImageData,
        help='Image data. See `image_data.ImageData`.'
    )

    @t.default('color_mode_data')
    def _default_color_mode_data(self):
        return color_mode.ColorModeData()

    @t.default('image_resources')
    def _default_image_resources(self):
        return image_resources.ImageResources()

    @t.default('layer_and_mask_info')
    def _default_layer_and_mask_info(self):
        return layers.LayerAndMaskInfo()

    @t.default('image_data')
    def _default_image_data(self):
        return image_data.ImageData()

    @classmethod
    @util.trace_read
    def read(cls, fd):
        self = cls.header_read(fd)
        self.color_mode_data = color_mode.ColorModeData.read(fd, self)
        self.image_resources = image_resources.ImageResources.read(fd, self)
        self.layer_and_mask_info = layers.LayerAndMaskInfo.read(fd, self)
        self.image_data = image_data.ImageData.read(fd, self)
        return self
    read.__func__.__doc__ = docs.read_single

    @util.trace_write
    def write(self, fd):
        Header.write(self, fd)
        self.color_mode_data.write(fd, self)
        self.image_resources.write(fd, self)
        self.layer_and_mask_info.write(fd, self)
        self.image_data.write(fd, self)
    write.__doc__ = docs.write_single



# -*- coding: utf-8 -*-


"""
The `ImageData` section.
"""


from __future__ import unicode_literals, absolute_import


import struct


import traitlets as t


from . import codecs
from . import enums
from . import util


class ImageData(t.HasTraits):
    """
    Stores (non-layer) image data.
    """
    compression = t.Enum(
        list(enums.Compression),
        default_value=enums.Compression.rle,
        help="Compression method. See `enums.Compression`"
    )

    def __init__(self, channels=None, fd=None, offset=None, size=None,
                 height=None, width=None, num_channels=None, depth=None,
                 version=None, compression=enums.Compression.raw):
        t.HasTraits.__init__(self, compression=compression)
        if channels is not None:
            if (fd is not None or offset is not None or size is not None or
                    height is not None or width is not None or
                    num_channels is not None or depth is not None or
                    version is not None):
                raise ValueError(
                    "May not provide both channels and other parameters")
            self._validate_channels(channels)
            self._channels = channels
            self._fd = None
        elif fd is not None:
            if channels is not None:
                raise ValueError(
                    "May not provide both channels and other parameters")
            self._image = None
            self._fd = fd
            self._offset = offset
            self._size = size
            self._height = height
            self._width = width
            self._num_channels = num_channels
            self._depth = depth
            self._version = version
        else:
            self._channels = None
            self._fd = None

    def _validate_channels(self, channels):
        if channels is None:
            return
        if len(channels.shape) != 3:
            raise ValueError("image must be a 3-dimensional array")
        if channels.dtype.kind != 'u':
            raise ValueError("image must have unsigned integer data type")

    @property
    def channels(self):
        if self._fd is None:
            return self._channels

        tell = self._fd.tell()
        self._fd.seek(self._offset)
        try:
            data = self._fd.read(self._size)
            image = codecs.decompress_image(
                data, self.compression,
                (self._height * self._num_channels, self._width),
                self._depth, self._version)
            return image.reshape(
                (self._num_channels, self._height, self._width))
        finally:
            self._fd.seek(tell)

    @property
    def shape(self):
        if self._fd is None:
            return self._channels.shape
        return self._shape

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        compression = fd.read(2)
        if compression == b'':
            raise IOError("Unexpected end of file")

        compression = struct.unpack('>H', compression)[0]
        util.log("compression: {}", enums.Compression(compression))

        offset = fd.tell()
        fd.seek(0, 2)
        size = fd.tell() - offset

        return cls(fd=fd, offset=offset, size=size, height=header.height,
                   width=header.width, num_channels=header.num_channels,
                   depth=header.depth, version=header.version,
                   compression=compression)

    @util.trace_write
    def write(self, fd, header):
        util.write_value(fd, 'H', self.compression)

        if self._fd is None:
            if self.channels is None:
                channels = 0
            else:
                channels = self.channels
            codecs.compress_image(
                fd, channels, self.compression, header.shape,
                header.num_channels, header.depth, header.version)
        else:
            tell = self._fd.tell()
            try:
                self._fd.seek(self._offset)
                data = self._fd.read(self._size)
            finally:
                self._fd.seek(tell)
            fd.write(data)



# -*- coding: utf-8 -*-


"""
Manage blending ranges.
"""


from __future__ import unicode_literals, absolute_import


import struct


import traitlets as t


from . import docs
from . import util


class BlendingRange(t.HasTraits):
    """
    Blending range data.

    Comprises 2 black values and 2 white values.
    """
    black0 = t.Int(min=0, max=255)
    black1 = t.Int(min=0, max=255)
    white0 = t.Int(min=0, max=255)
    white1 = t.Int(min=0, max=255)

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        black0, black1, white0, white1 = struct.unpack('>BBBB', fd.read(4))

        util.log(
            "black: ({}, {}), white: ({}, {})",
            black0, black1, white0, white1)

        return cls(
            black0=black0,
            black1=black1,
            white0=white0,
            white1=white1)
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        fd.write(struct.pack(
            '>BBBB', self.black0, self.black1, self.white0, self.white1))
    write.__doc__ = docs.write


class BlendingRangePair(t.HasTraits):
    """
    Blending range pair.

    The combination of a source and destination blending range.
    """
    src = t.Instance(
        BlendingRange,
        help="Source `BlendingRange`"
    )
    dst = t.Instance(
        BlendingRange,
        help="Destination `BlendingRange`"
    )

    @t.default('src')
    def _default_src(self):
        return BlendingRange()

    @t.default('dst')
    def _default_dst(self):
        return BlendingRange()

    def length(self, header):
        return 8
    length.__doc__ = docs.length

    def total_length(self, header):
        return self.length(header)
    total_length.__doc__ = docs.total_length

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        src = BlendingRange.read(fd, header)
        dst = BlendingRange.read(fd, header)

        return cls(src=src,
                   dst=dst)
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        self.src.write(fd, header)
        self.dst.write(fd, header)
    write.__doc__ = docs.write


class BlendingRanges(t.HasTraits):
    """
    All of the layer blending range data.

    Consists of a composite gray blend pair followed by N additional
    pairs.
    """
    composite_gray_blend = t.Instance(
        BlendingRangePair, allow_none=True,
        help="Composite gray `BlendingRangePair`."
    )
    channels = t.List(
        t.Instance(BlendingRangePair),
        help="List of additional `BlendingRangePair` instances."
    )

    def length(self, header):
        if (self.composite_gray_blend is not None or
                len(self.channels)):
            if self.composite_gray_blend is None:
                composite_gray_blend = BlendingRangePair()
            else:
                composite_gray_blend = self.composite_gray_blend
            return (
                composite_gray_blend.total_length(header) +
                sum(x.total_length(header) for x in self.channels))
        return 0
    length.__doc__ = docs.length

    def total_length(self, header):
        return 4 + self.length(header)
    total_length.__doc__ = docs.total_length

    @classmethod
    @util.trace_read
    def read(cls, fd, header, num_channels):
        length = util.read_value(fd, 'I')
        end = fd.tell() + length
        util.log("length: {}, end: {}", length, end)
        if length == 0:
            return cls()

        composite_gray_blend = BlendingRangePair.read(fd, header)
        channels = []
        while fd.tell() < end:
            channels.append(BlendingRangePair.read(fd, header))

        fd.seek(end)

        return cls(
            composite_gray_blend=composite_gray_blend,
            channels=channels)
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        util.write_value(fd, 'I', self.length(header))
        if (self.composite_gray_blend is not None or
                len(self.channels)):
            if self.composite_gray_blend is None:
                composite_gray_blend = BlendingRangePair()
            else:
                composite_gray_blend = self.composite_gray_blend
            composite_gray_blend.write(fd, header)
            for channel in self.channels:
                channel.write(fd, header)
    write.__doc__ = docs.write



# -*- coding: utf-8 -*-


"""
Handle Bézier paths.
"""


from __future__ import unicode_literals, absolute_import


import struct


import six


import traitlets as t


from . import docs
from . import enums
from . import util


def _read_point(x, size):
    return (float(x) / (1 << 24)) * float(size)


def _write_point(x, size):
    return int((x / float(size)) * (1 << 24))


class _PathRecordMeta(type(t.HasTraits)):
    """
    A metaclass that builds a mapping of subclasses.
    """
    mapping = {}

    def __new__(cls, name, parents, dct):
        new_cls = type(t.HasTraits).__new__(cls, name, parents, dct)

        if 'type' in dct:
            if dct['type'] in cls.mapping:
                raise ValueError("Duplicate type '{}'".format(dct['type']))
            cls.mapping[dct['type']] = new_cls

        return new_cls


@six.add_metaclass(_PathRecordMeta)
class PathRecord(t.HasTraits):
    def length(self, header):
        return 26

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        type = util.read_value(fd, 'H')

        new_cls = _PathRecordMeta.mapping[type]

        return new_cls.read_data(fd, header)

    def write(self, fd, header):
        util.write_value(fd, 'H', self.type)
        self.write_data(fd, header)


class PathFillRuleRecord(PathRecord):
    type = enums.PathRecordType.path_fill_rule_record

    @classmethod
    @util.trace_read
    def read_data(cls, fd, header):
        padding = fd.read(24)
        if padding != b'\0' * 24:  # pragma: no cover
            raise ValueError(
                "Invalid padding in path fill rule record")
        return cls()

    def write_data(self, fd, header):
        fd.write(b'\0' * 24)


class InitialFillRuleRecord(PathRecord):
    type = enums.PathRecordType.initial_fill_rule_record
    all_pixels = t.Bool(
        help='Fill starts with all pixels'
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, header):
        all_pixels = bool(util.read_value(fd, 'H'))
        padding = fd.read(22)
        if padding != b'\0' * 22:  # pragma: no cover
            raise ValueError(
                "Invalid padding in initial fill rule record")

        util.log("all_pixels: {}", all_pixels)

        return cls(all_pixels=all_pixels)

    def write_data(self, fd, header):
        util.write_value(fd, 'H', self.all_pixels)
        fd.write(b'\0' * 22)


class _LengthRecord(PathRecord):
    num_knots = t.Int(
        help="Number of Bezier knots"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, header):
        num_knots = util.read_value(fd, 'H')
        fd.read(22)
        util.log('num_knots: {}', num_knots)
        return cls(num_knots=num_knots)

    def write_data(self, fd, header):
        util.write_value(fd, 'H', self.num_knots)
        fd.write(b'\0' * 22)


class ClosedSubpathLengthRecord(_LengthRecord):
    type = enums.PathRecordType.closed_subpath_length


class OpenSubpathLengthRecord(_LengthRecord):
    type = enums.PathRecordType.open_subpath_length


class _PointRecord(PathRecord):
    y0 = t.Float(
        help='y of control point preceding the knot, in pixels'
    )
    x0 = t.Float(
        help='x of control point preceding the knot, in pixels'
    )
    y1 = t.Float(
        None,
        allow_none=True,
        help='y of anchor point of the knot'
    )
    x1 = t.Float(
        None,
        allow_none=True,
        help='x of anchor point of the knot'
    )
    y2 = t.Float(
        None,
        allow_none=True,
        help='y of control point for the segment leaving the knot, in pixels'
    )
    x2 = t.Float(
        None,
        allow_none=True,
        help='x of control point for the segment leaving the knot, in pixels'
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, header):
        data = fd.read(24)
        y0, x0, y1, x1, y2, x2 = struct.unpack('>iiiiii', data)

        y0 = _read_point(y0, header.height)
        x0 = _read_point(x0, header.width)
        y1 = _read_point(y1, header.height)
        x1 = _read_point(x1, header.width)
        y2 = _read_point(y2, header.height)
        x2 = _read_point(x2, header.width)

        util.log(
            '({}, {}) ({}, {}), ({}, {})',
            y0, x0, y1, x1, y2, x2)

        return cls(y0=y0, x0=x0, y1=y1, x1=x1, y2=y2, x2=x2)

    def write_data(self, fd, header):
        y0 = _write_point(self.y0, header.height)
        x0 = _write_point(self.x0, header.width)

        if self.y1 is None:
            y1 = y0
        else:
            y1 = _write_point(self.y1, header.height)
        if self.x1 is None:
            x1 = x0
        else:
            x1 = _write_point(self.x1, header.width)

        if self.y2 is None:
            y2 = y0
        else:
            y2 = _write_point(self.y2, header.height)
        if self.x2 is None:
            x2 = x0
        else:
            x2 = _write_point(self.x2, header.width)

        fd.write(struct.pack('>iiiiii', y0, x0, y1, x1, y2, x2))


class ClosedSubpathBezierKnotLinked(_PointRecord):
    type = enums.PathRecordType.closed_subpath_bezier_knot_linked


class ClosedSubpathBezierKnotUnlinked(_PointRecord):
    type = enums.PathRecordType.closed_subpath_bezier_knot_unlinked


class OpenSubpathBezierKnotLinked(_PointRecord):
    type = enums.PathRecordType.open_subpath_bezier_knot_linked


class OpenSubpathBezierKnotUnlinked(_PointRecord):
    type = enums.PathRecordType.open_subpath_bezier_knot_unlinked


class ClipboardRecord(PathRecord):
    type = enums.PathRecordType.clipboard_record

    top = t.Float(
        help="top, in pixels"
    )
    left = t.Float(
        help="left, in pixels"
    )
    bottom = t.Float(
        help="bottom, in pixels"
    )
    right = t.Float(
        help="right, in pixels"
    )
    resolution = t.Int(
        help="resolution"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, header):
        data = fd.read(24)
        top, left, bottom, right, resolution, _ = struct.unpack(
            '>iiiiii', data)

        top = _read_point(top, header.height)
        left = _read_point(left, header.width)
        bottom = _read_point(bottom, header.height)
        right = _read_point(right, header.width)

        util.log(
            'position: ({}, {}, {}, {}), resolution: {}',
            top, left, bottom, right, resolution)

        return cls(top=top, left=left, bottom=bottom, right=right,
                   resolution=resolution)

    def write_data(self, fd, header):
        top = _write_point(self.top, header.height)
        left = _write_point(self.left, header.width)
        bottom = _write_point(self.bottom, header.height)
        right = _write_point(self.right, header.width)

        fd.write(
            struct.pack('>iiiiii',
                        top, left, bottom, right, self.resolution, 0))


class PathResource(t.HasTraits):
    path_records = t.List(
        t.Instance(PathRecord),
        help="List of `PathRecord` instances."
    )

    def length(self, header):
        return sum(x.length(header) for x in self.path_records)
    length.__doc__ = docs.length

    @classmethod
    @util.trace_read
    def read(cls, fd, length, header):
        end = fd.tell() + length

        path_records = []
        while fd.tell() + 26 <= end:
            path_records.append(
                PathRecord.read(fd, header))

        if (path_records[0].type !=
                enums.PathRecordType.path_fill_rule_record):
            raise ValueError(
                'Path resource must start with path_fill_rule_record')

        fd.seek(end - fd.tell(), 1)

        return cls(path_records=path_records)

    @classmethod
    def from_rect(cls, top, left, bottom, right, all_pixels=True):
        return cls(path_records=[
            PathFillRuleRecord(),
            InitialFillRuleRecord(all_pixels=False),
            OpenSubpathLengthRecord(num_knots=5),
            OpenSubpathBezierKnotLinked(y0=top, x0=left),
            OpenSubpathBezierKnotLinked(y0=top, x0=right),
            OpenSubpathBezierKnotLinked(y0=bottom, x0=right),
            OpenSubpathBezierKnotLinked(y0=bottom, x0=left),
            OpenSubpathBezierKnotLinked(y0=top, x0=left),
        ]
        )

    def write(self, fd, header):
        for path_record in self.path_records:
            path_record.write(fd, header)



# -*- coding: utf-8 -*-


"""
The `ColorModeData` section.
"""


from __future__ import unicode_literals, absolute_import


import traitlets as t


from . import docs
from . import util


class ColorModeData(t.HasTraits):
    """
    Color mode data section.

    Only indexed color and duotone (see `core.Header.color_mode`) have
    color mode data.

    Indexed color images: length is 768; color data contains the color
    table for the image, in non-interleaved order.

    Duotone images: color data contains the duotone specification (the
    format of which is not documented). Other applications that read
    Photoshop files can treat a duotone image as a gray image, and
    just preserve the contents of the duotone information when reading
    and writing the file.

    Note that ``pytoshop`` doesn't do anything meaningful for color
    mode data, and only stores the raw bytes in order to round-trip.
    """
    data = t.Bytes(
        help='The raw color mode data.'
    )

    def length(self, header):
        return len(self.data)
    length.__doc__ = docs.length

    @classmethod
    @util.trace_read
    def read(cls, fd, header):
        length = util.read_value(fd, 'I')
        util.log("length: {}", length)
        data = fd.read(length)
        return cls(data=data)
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header):
        util.write_value(fd, 'I', self.length(header))
        fd.write(self.data)
    write.__doc__ = docs.write



# -*- coding: utf-8 -*-


import io


import numpy as np
from numpy.testing import assert_array_equal
import pytest


from pytoshop import codecs
from pytoshop import enums


@pytest.mark.parametrize("depth", (8, 16))
def test_zip_with_prediction(depth):
    np.random.seed(0)

    dtype = codecs.color_depth_dtype_map[depth]
    x = np.random.randint(0, (2**depth) - 1, size=(255, 256), dtype=dtype)

    fd = io.BytesIO()
    codecs.compress_image(
        fd, x, enums.Compression.zip_prediction, (255, 256), 1, depth, 1)
    y = codecs.decompress_image(
        fd.getvalue(), enums.Compression.zip_prediction, (255, 256), depth, 1)

    assert_array_equal(x, y)


@pytest.mark.parametrize("depth", (1, 8, 16, 32))
def test_zip(depth):
    np.random.seed(0)

    dtype = codecs.color_depth_dtype_map[depth]
    x = np.random.randint(0, (2**depth) - 1, size=(255, 256), dtype=dtype)

    fd = io.BytesIO()
    codecs.compress_image(
        fd, x, enums.Compression.zip, (255, 256), 1, depth, 1)
    y = codecs.decompress_image(
        fd.getvalue(), enums.Compression.zip, (255, 256), depth, 1)

    assert_array_equal(x, y)


@pytest.mark.parametrize("depth", (8, 16, 32))
@pytest.mark.parametrize("version", (1, 2))
def test_rle(depth, version):
    np.random.seed(0)

    dtype = codecs.color_depth_dtype_map[depth]
    x = np.random.randint(0, (2**depth) - 1, size=(255, 256), dtype=dtype)

    fd = io.BytesIO()
    codecs.compress_image(
        fd, x, enums.Compression.rle, (255, 256), 1, depth, version)
    y = codecs.decompress_image(
        fd.getvalue(), enums.Compression.rle, (255, 256), depth, version)

    assert_array_equal(x, y)


@pytest.mark.parametrize("depth", (1, 8, 16, 32))
def test_raw_constant(depth):
    if depth == 1:
        value = 1
    else:
        value = 42

    dtype = codecs.color_depth_dtype_map[depth]
    x = np.ones((255, 256), dtype=dtype) * value

    fd = io.BytesIO()
    codecs.compress_image(
        fd, value, enums.Compression.raw, (255, 256), 1, depth, 1)
    y = codecs.decompress_image(
        fd.getvalue(), enums.Compression.raw, (255, 256), depth, 1)

    assert_array_equal(x, y)


@pytest.mark.parametrize("depth", (8, 16))
def test_zip_with_prediction_constant(depth):
    dtype = codecs.color_depth_dtype_map[depth]
    x = np.ones((255, 256), dtype=dtype) * 42

    fd = io.BytesIO()
    codecs.compress_image(
        fd, 42, enums.Compression.zip_prediction, (255, 256), 1, depth, 1)
    y = codecs.decompress_image(
        fd.getvalue(), enums.Compression.zip_prediction, (255, 256), depth, 1)

    assert_array_equal(x, y)


@pytest.mark.parametrize("depth", (1, 8, 16, 32))
def test_zip_constant(depth):
    if depth == 1:
        value = 1
    else:
        value = 42

    dtype = codecs.color_depth_dtype_map[depth]
    x = np.ones((255, 256), dtype=dtype) * value

    fd = io.BytesIO()
    codecs.compress_image(
        fd, value, enums.Compression.zip, (255, 256), 1, depth, 1)
    y = codecs.decompress_image(
        fd.getvalue(), enums.Compression.zip, (255, 256), depth, 1)

    assert_array_equal(x, y)


@pytest.mark.parametrize("depth", (8, 16, 32))
@pytest.mark.parametrize("version", (1, 2))
def test_rle_constant(depth, version):
    dtype = codecs.color_depth_dtype_map[depth]
    x = np.ones((255, 256), dtype=dtype) * 42

    fd = io.BytesIO()
    codecs.compress_image(
        fd, 42, enums.Compression.rle, (255, 256), 1, depth, version)
    y = codecs.decompress_image(
        fd.getvalue(), enums.Compression.rle, (255, 256), depth, version)

    assert_array_equal(x, y)



# -*- coding: utf-8 -*-


from __future__ import unicode_literals, absolute_import


from .core import PsdFile


__author__ = 'Michael Droettboom'
__email__ = 'mdboom@gmail.com'
__version__ = '0.3.0'


def read(fd):
    """
    Read a PSD file from a file-like object.

    Parameters
    ----------
    fd : file-like object
        Must be readable, seekable and open in binary mode.

    Returns
    -------
    psdfile : PsdFile
    """
    return PsdFile.read(fd)



# -*- coding: utf-8 -*-


"""
`TaggedBlock` objects.
"""


from __future__ import unicode_literals, absolute_import


import six


import traitlets as t


from . import docs
from . import enums
from . import path
from . import util


class _TaggedBlockMeta(type(t.HasTraits)):
    """
    A metaclass that builds a mapping of subclasses.
    """
    mapping = {}

    def __new__(cls, name, parents, dct):
        new_cls = type(t.HasTraits).__new__(cls, name, parents, dct)

        if 'code' in dct and isinstance(dct['code'], bytes):
            if dct['code'] in cls.mapping:
                raise ValueError("Duplicate code '{}'".format(dct['code']))
            cls.mapping[dct['code']] = new_cls

        return new_cls


@six.add_metaclass(_TaggedBlockMeta)
class TaggedBlock(t.HasTraits):
    _large_layer_info_codes = set([
        b'LMsk', b'Lr16', b'Lr32', b'Layr', b'Mt16', b'Mt32',
        b'Mtrn', b'Alph', b'FMsk', b'Ink2', b'FEid', b'FXid',
        b'PxSD'])

    def length(self, header):
        return self.data_length(header)
    length.__doc__ = docs.length

    def total_length(self, header, padding=1):
        length = 8
        if header.version == 2 and self.code in self._large_layer_info_codes:
            length += 8
        else:
            length += 4
        length += util.pad(self.length(header), padding)
        return length
    total_length.__doc__ = docs.total_length

    @classmethod
    @util.trace_read
    def read(cls, fd, header, padding=1):
        signature = fd.read(4)
        if signature not in (b'8BIM', b'8B64'):
            raise ValueError('Invalid signature in tagged block')

        code = fd.read(4)

        if header.version == 2 and code in cls._large_layer_info_codes:
            length = util.read_value(fd, 'Q')
        else:
            length = util.read_value(fd, 'I')
        padded_length = util.pad(length, padding)

        util.log(
            "code: {}, length: {}, padded_length: {}",
            code, length, padded_length
        )

        new_cls = _TaggedBlockMeta.mapping.get(code, GenericTaggedBlock)
        start = fd.tell()
        result = new_cls.read_data(fd, code, length, header)
        end = fd.tell()
        if end - start != length:
            raise ValueError("{} read the wrong amount".format(new_cls))
        fd.seek(padded_length - length, 1)

        return result
    read.__func__.__doc__ = docs.read

    @util.trace_write
    def write(self, fd, header, padding=1):
        if header.version == 2 and self.code in self._large_layer_info_codes:
            fd.write(b'8B64')
        else:
            fd.write(b'8BIM')
        fd.write(self.code)

        length = self.data_length(header)
        padded_length = util.pad(length, padding)
        if header.version == 2 and self.code in self._large_layer_info_codes:
            util.write_value(fd, 'Q', length)
        else:
            util.write_value(fd, 'I', length)

        start = fd.tell()
        self.write_data(fd, header)
        end = fd.tell()
        if end - start != length:
            raise ValueError(
                "{} wrote the wrong amount".format(self.__class__))
        fd.write(b'\0' * (padded_length - length))
    write.__doc__ = docs.write


class GenericTaggedBlock(TaggedBlock):
    """
    A generic `TaggedBlock` subclass for tag codes ``pytoshop``
    doesn't know about.
    """

    code = t.Bytes(
        min=4, max=4,
        help="The 4-letter tagged block code"
    )
    data = t.Bytes(
        b'',
        help="The raw data for the block"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, code, length, header):
        data = fd.read(length)
        return cls(code=code, data=data)

    def data_length(self, header):
        return len(self.data)

    @util.trace_write
    def write_data(self, fd, header):
        fd.write(self.data)


class UnicodeLayerName(TaggedBlock):
    code = b'luni'
    name = t.Unicode(
        help="The name of the layer."
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, code, length, header):
        data = fd.read(length)
        name = util.decode_unicode_string(data)
        return cls(name=name)

    def data_length(self, header):
        return len(util.encode_unicode_string(self.name))

    @util.trace_write
    def write_data(self, fd, header):
        fd.write(util.encode_unicode_string(self.name))


class LayerId(TaggedBlock):
    code = b'lyid'
    id = t.Int(
        help="Layer id"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, code, length, header):
        id = util.read_value(fd, 'I')
        return cls(id=id)

    def data_length(self, header):
        return 4

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'I', self.id)


class LayerNameSource(TaggedBlock):
    code = b'lnsr'
    id = t.Int(
        help="The layer id of the source of the name of this layer"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, code, length, header):
        id = util.read_value(fd, 'I')
        return cls(id=id)

    def data_length(self, header):
        return 4

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'I', self.id)


class _SectionDividerSetting(TaggedBlock):
    type = t.Enum(
        list(enums.SectionDividerSetting),
        default_value=enums.SectionDividerSetting.open,
        help="Section divider type. See `enums.SectionDividerSetting`."
    )
    key = t.Enum(
        list(enums.BlendMode), allow_none=True,
        help="Section divider key"
    )
    subtype = t.Bool(
        None, allow_none=True,
        help="Section divider subtype. False=normal, True=Scene group, "
        "affects the animation timeline"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, code, length, header):
        end = fd.tell() + length
        type = util.read_value(fd, 'I')
        key = None
        subtype = None
        if fd.tell() < end:
            sig = fd.read(4)
            if sig != b'8BIM':
                raise ValueError("Invalid signature")
            key = fd.read(4)
            if fd.tell() < end:
                subtype = bool(util.read_value(fd, 'I'))

        return cls(type=type, key=key, subtype=subtype)

    def data_length(self, header):
        length = 4
        if self.subtype is not None:
            length += 12
        elif self.key is not None:
            length += 8
        return length

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'I', self.type)
        if self.key is not None or self.subtype is not None:
            if self.key is None:
                key = b'norm'
            else:
                key = self.key
            fd.write(b'8BIM')
            fd.write(key)
            if self.subtype is not None:
                util.write_value(fd, 'I', self.subtype)


class SectionDividerSetting(_SectionDividerSetting):
    code = b'lsct'


class NestedSectionDividerSetting(_SectionDividerSetting):
    code = b'lsdk'


class VectorMask(TaggedBlock):
    code = b'vmsk'
    version = t.Int(
        3,
        help='Vector mask block version'
    )
    invert = t.Bool(
        help='Invert mask'
    )
    not_link = t.Bool(
        help="Don't link mask"
    )
    disable = t.Bool(
        help="Disable mask"
    )
    path_resource = t.Instance(
        path.PathResource,
        help="`path.PathResource` instance`"
    )

    @classmethod
    @util.trace_read
    def read_data(cls, fd, code, length, header):
        version = util.read_value(fd, 'I')
        flags = util.read_value(fd, 'I')
        invert, not_link, disable = util.unpack_bitflags(flags, 3)

        util.log(
            "version: {}, invert: {}, not_link: {}, disable: {}",
            version, invert, not_link, disable)

        path_resource = path.PathResource.read(fd, length - 8, header)

        return cls(
            version=version,
            invert=invert,
            not_link=not_link,
            disable=disable,
            path_resource=path_resource)

    def data_length(self, header):
        return 8 + self.path_resource.length(header)

    @util.trace_write
    def write_data(self, fd, header):
        util.write_value(fd, 'I', self.version)

        flags = util.pack_bitflags(self.invert, self.not_link, self.disable)

        util.write_value(fd, 'I', flags)

        self.path_resource.write(fd, header)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import io


import pytest


import traitlets as t


from pytoshop import core
from pytoshop import enums


class TestHeader(object):
    def test_header(self):
        content = b'8BPS\0\x02\0\0\0\0\0\0\0\x03\0\0\0\x0F\0\0\0\x0F\0\x08\0\1'
        fd = io.BytesIO(content)
        fd.seek(0)
        h = core.Header.read(fd)

        assert h.version == 2
        assert h.num_channels == 3
        assert h.height == 15
        assert h.width == 15
        assert h.depth == 8
        assert h.color_mode == enums.ColorMode.grayscale

        fd = io.BytesIO()
        h.write(fd)
        assert fd.getvalue() == content

    def test_header_invalid_version(self):
        content = b'8BPS\0\x03\0\0\0\0\0\0\0\x03\0\0\0\x0F\0\0\0\x0F\0\x08\0\1'
        fd = io.BytesIO(content)
        fd.seek(0)
        with pytest.raises(ValueError):
            core.Header.read(fd)

    def test_header_invalid_signature(self):
        content = b'8BPX\0\x02\0\0\0\0\0\0\0\x03\0\0\0\x0F\0\0\0\x0F\0\x08\0\1'
        fd = io.BytesIO(content)
        fd.seek(0)
        with pytest.raises(ValueError):
            core.Header.read(fd)

    def test_header_invalid_width(self):
        content = b'8BPS\0\x02\0\0\0\0\0\0\0\x03\0\0\0\0\0\0\0\x0F\0\x08\0\1'
        fd = io.BytesIO(content)
        fd.seek(0)
        with pytest.raises(t.TraitError):
            core.Header.read(fd)

    def test_header_invalid_depth(self):
        content = b'8BPS\0\x02\0\0\0\0\0\0\0\x03\0\0\0\x0F\0\0\0\x0F\0\x09\0\1'
        fd = io.BytesIO(content)
        fd.seek(0)
        with pytest.raises(t.TraitError):
            core.Header.read(fd)

    def test_size_too_large(self):
        with pytest.raises(t.TraitError):
            core.Header(version=1, width=30001, height=5)
        core.Header(version=2, width=30001, height=5)

        with pytest.raises(t.TraitError):
            core.Header(version=2, width=300001, height=5)

        with pytest.raises(t.TraitError):
            core.Header(version=1, width=5, height=30001)
        core.Header(version=2, width=5, height=30001)

        with pytest.raises(t.TraitError):
            core.Header(version=2, width=5, height=300001)



# -*- coding: utf-8 -*-


import io
import os


import numpy as np
import pytest


import pytoshop
from pytoshop import enums
from pytoshop.user import nested_layers


DATA_PATH = os.path.join(os.path.dirname(__file__), 'psd_files')


def test_nested_layers():
    filename = os.path.join(DATA_PATH, 'group.psd')
    with open(filename, 'rb') as fd:
        psd = pytoshop.PsdFile.read(fd)

        layers = nested_layers.psd_to_nested_layers(psd)

        nested_layers.pprint_layers(layers)

        psd2 = nested_layers.nested_layers_to_psd(layers, enums.ColorMode.rgb)

        fd = io.BytesIO()
        psd2.write(fd)


def test_nested_layers_no_adjust():
    filename = os.path.join(DATA_PATH, 'group.psd')
    with open(filename, 'rb') as fd:
        psd = pytoshop.PsdFile.read(fd)

        layers = nested_layers.psd_to_nested_layers(psd)

        psd2 = nested_layers.nested_layers_to_psd(
            layers, enums.ColorMode.rgb, size=(psd.width, psd.height))

        fd = io.BytesIO()
        psd2.write(fd)


def test_errors():
    with pytest.raises(TypeError):
        nested_layers.psd_to_nested_layers(None)


@pytest.mark.parametrize("vector_mask", (True, False))
def test_from_scratch(vector_mask):
    from pytoshop.user.nested_layers import Group, Image

    img1 = np.empty((100, 80), dtype=np.uint8)
    img2 = np.empty((100, 80), dtype=np.uint8)
    img3 = np.empty((100, 80), dtype=np.uint8)
    img4 = np.empty((2, 100, 80), dtype=np.uint8)

    layers = [
        Group(
            layers=[
                Image(channels={0: img1},
                      top=0, left=0, bottom=100, right=80),
                Image(channels=img2,
                      top=15, left=15),
                Image(channels=img4,
                      top=42, left=43),
                Image(channels=[img1, img1],
                      top=-5, left=49)
            ]),
        Image(channels={0: img3})
    ]

    psd = nested_layers.nested_layers_to_psd(
        layers, enums.ColorMode.grayscale,
        vector_mask=vector_mask)

    buff = io.BytesIO()
    psd.write(buff)

    buff.seek(0)

    pytoshop.read(buff)


def test_mixed_depth():
    from pytoshop.user.nested_layers import Group, Image

    img1 = np.empty((100, 80), dtype=np.uint8)
    img2 = np.empty((100, 80), dtype=np.uint16)

    layers = [
        Group(
            layers=[
                Image(channels={0: img1},
                      top=0, left=0, bottom=100, right=80),
                Image(channels=img2,
                      top=15, left=15),
            ])
    ]

    with pytest.raises(ValueError):
        nested_layers.nested_layers_to_psd(
            layers, enums.ColorMode.grayscale)


def test_mismatched_height():
    from pytoshop.user.nested_layers import Group, Image

    img1 = np.empty((100, 80), dtype=np.uint8)

    layers = [
        Group(
            layers=[
                Image(channels={0: img1},
                      top=0, left=0, bottom=101, right=80),
            ])
    ]

    with pytest.raises(ValueError):
        nested_layers.nested_layers_to_psd(
            layers, enums.ColorMode.grayscale)


def test_mismatched_width():
    from pytoshop.user.nested_layers import Group, Image

    img1 = np.empty((100, 80), dtype=np.uint8)

    layers = [
        Group(
            layers=[
                Image(channels={0: img1},
                      top=0, left=0, bottom=100, right=81),
            ])
    ]

    with pytest.raises(ValueError):
        nested_layers.nested_layers_to_psd(
            layers, enums.ColorMode.grayscale)


def test_no_images():
    from pytoshop.user.nested_layers import Group

    layers = [
        Group(
            layers=[]
            )
        ]

    with pytest.raises(ValueError):
        nested_layers.nested_layers_to_psd(
            layers, enums.ColorMode.grayscale)


@pytest.mark.parametrize("compression", (0, 1, 2, 3))
def test_proxy(compression):
    from pytoshop.user.nested_layers import Group, Image

    class ImageProxy:
        @property
        def shape(self):
            return (1000, 1000)

        @property
        def dtype(self):
            return np.uint8().dtype

        def __array__(self):
            return np.arange(1000000, dtype=np.uint8).reshape((1000, 1000))

    image_layers = []
    for i in range(256):
        image_layers.append(
            Image(
                channels={0: ImageProxy()},
                top=i, left=i
            )
        )

    layers = [
        Group(
            layers=image_layers
        )
    ]

    psd = nested_layers.nested_layers_to_psd(
        layers, enums.ColorMode.grayscale,
        compression=compression
    )

    buff = io.BytesIO()
    psd.write(buff)

    buff.seek(0)

    psd = pytoshop.read(buff)
    layers = nested_layers.psd_to_nested_layers(psd)
    # Extract all the data so it's tested and included in the timings
    for image in layers[0].layers:
        [x.image for x in image.channels.values()]


if __name__ == '__main__':
    test_proxy(1)



# -*- coding: utf-8 -*-


import io
import os


import numpy as np
import pytest
import traitlets as t


import pytoshop
from pytoshop import layers


DATA_PATH = os.path.join(os.path.dirname(__file__), 'psd_files')


def test_futz_with_channel_image_data():
    filename = os.path.join(DATA_PATH, 'group.psd')
    with open(filename, 'rb') as fd:
        psd = pytoshop.PsdFile.read(fd)

        first_layer = psd.layer_and_mask_info.layer_info.layer_records[0]

        first_layer.channels[0].image = np.empty((256, 256))
        with pytest.raises(ValueError):
            psd.write(io.BytesIO())

        first_layer.channels[0].image = np.empty((256,), np.uint8)
        with pytest.raises(ValueError):
            psd.write(io.BytesIO())

        first_layer.channels[0].image = np.empty((256, 256), np.uint8)
        with pytest.raises(ValueError):
            psd.write(io.BytesIO())

        first_layer.channels[0].image = 0
        psd.write(io.BytesIO())

        first_layer.channels[0].image = np.empty((200, 100), np.uint8)
        psd.write(io.BytesIO())


def test_futz_with_layer_channels():
    filename = os.path.join(DATA_PATH, 'group.psd')
    with open(filename, 'rb') as fd:
        psd = pytoshop.PsdFile.read(fd)

        first_layer = psd.layer_and_mask_info.layer_info.layer_records[0]

        first_layer.channels = {
            0: layers.ChannelImageData(image=np.empty((200, 100), np.uint8))}
        psd.write(io.BytesIO())

        with pytest.raises(t.TraitError):
            first_layer.channels = {
                0: np.empty((200, 100), np.uint8)}

        with pytest.raises(t.TraitError):
            first_layer.channels = {
                'zero': layers.ChannelImageData(
                    image=np.empty((200, 100), np.uint8))}



# -*- coding: utf-8 -*-


"""
Enumerated values used throughout the library.
"""


from __future__ import unicode_literals, absolute_import


import enum


class Version(enum.IntEnum):
    """
    The PSD file version.

    Version 1 is the classic "PSD" file.

    Version 2 is the large document format "PSB" file which supports
    documents up to 300,000 pixels in any dimension.
    """
    version_1 = 1
    version_2 = 2
    psd = 1
    psb = 2


class ColorDepth(enum.IntEnum):
    """
    Color depth (bits-per-pixel-per-channel).

    Supported values are 1, 8, 16, and 32.
    """
    depth1 = 1
    depth8 = 8
    depth16 = 16
    depth32 = 32


class ColorMode(enum.IntEnum):
    """
    Color mode.
    """
    bitmap = 0
    grayscale = 1
    indexed = 2
    rgb = 3
    cmyk = 4
    multichannel = 7
    duotone = 8
    lab = 9


class ColorSpace(enum.IntEnum):
    """
    Color space ids.

    Defines the meaning of the 4 color values in a color structure.

    - ``rgb``: The first three values are *red*, *green*, and *blue*.
      The are full 16-bit unsigned values as in Apple's ``RGBColor``
      data structure.

    - ``hsb``: The first three values are *hue*, *saturation*, and
      *brightness*.  The are full unsigned 16-bit values as in Apple's
      ``HSVColor`` data structure.

    - ``cmyk``: The four values are *cyan*, *magenta*, *yellow* and
      *black*.  The are full unsigned 16-bit values.  0 = 100% ink.

    - ``lab``: The first three values are *lightness*, *a
      chrominance*, and *b chrominance*.  Lightness is a 16-bit value
      from 0 to 10000.  Chrominance components are each 16-bit values
      from -12800 to 12700.  Gray values are represented by
      chrominance components of 0.

    - ``grayscale``: The first value in the color data is the gray
      value from 0 to 10000.

    Additional values are for "custom color spaces" which are not as
    well documented:

    - ``pantone``: Pantone® matching system.

    - ``focoltone``: Focoltone matching system.

    - ``trumatch``: Trumatch color.

    - ``toyo_88_colorfinder_1050``: Toyo 88 colorfinder 1050.

    - ``hks``: HKS colors.
    """
    rgb = 0
    hsb = 1
    cmyk = 2
    pantone = 3
    focoltone = 4
    trumatch = 5
    toyo_88_colorfinder_1050 = 6
    lab = 7
    grayscale = 8
    hks = 10


class BlendMode(bytes, enum.Enum):
    """
    Layer blend mode.
    """
    pass_through = b'pass'
    normal = b'norm'
    dissolve = b'diss'
    darken = b'dark'
    multiply = b'mul '
    color_burn = b'idiv'
    linear_burn = b'lbrn'
    darker_color = b'dkCl'
    lighten = b'lite'
    screen = b'scrn'
    color_dodge = b'div '
    linear_dodge = b'lddg'
    lighter_color = b'lgCl'
    overlay = b'over'
    soft_light = b'sLit'
    hard_light = b'hLit'
    vivid_light = b'vLit'
    linear_light = b'lLit'
    pin_light = b'pLit'
    hard_mix = b'hMix'
    difference = b'diff'
    exclusion = b'smud'
    subtract = b'fsub'
    divide = b'fdiv'
    hue = b'hue '
    saturation = b'sat '
    color = b'colr'
    luminosity = b'lum '


class Compression(enum.IntEnum):
    """
    Compression mode.

    - ``raw``: raw image data.

    - ``rle``: Run length encoded (RLE) compressed.  The RLE
      compression is the same compression algorithm used by the
      Macintosh ROM routine PackBits, and the TIFF standard.

    - ``zip``: Zip (zlib) without prediction.

    - ``zip_prediction``: Zip (zlib) with prediction.
    """
    raw = 0
    rle = 1
    zip = 2
    zip_prediction = 3


class LayerMaskKind(enum.IntEnum):
    """
    Layer mask kind.

    According to the spec, only ``use_value_stored_per_layer`` is
    preferred.  The other are retained for backward compatibility
    only.
    """
    color_selected = 0
    color_protected = 1
    use_value_stored_per_layer = 128


class ChannelId(enum.IntEnum):
    """
    Channel id.

    Used to map channel data to image planes in layers.

    The meaning of the positive numbers depends on the `ColorMode` in
    effect.
    """
    bitmap = 0

    gray = 0

    default = 0

    red = 0
    green = 1
    blue = 2

    cyan = 0
    magenta = 1
    yellow = 2
    black = 3

    L = 0
    a = 1
    b = 2

    transparency = -1
    user_layer_mask = -2
    real_user_layer_mask = -3


class SectionDividerSetting(enum.IntEnum):
    """
    Section divider setting.

    Used for the mode of grouped layers.

    - ``any_other``: any other type of layer

    - ``open``: Display an open folder

    - ``closed``: Display a closed folder

    - ``bounding``: Bounding section divider, hidden in GUI.
    """
    any_other = 0
    open = 1
    closed = 2
    bounding = 3


class PathRecordType(enum.IntEnum):
    closed_subpath_length = 0
    closed_subpath_bezier_knot_linked = 1
    closed_subpath_bezier_knot_unlinked = 2
    open_subpath_length = 3
    open_subpath_bezier_knot_linked = 4
    open_subpath_bezier_knot_unlinked = 5
    path_fill_rule_record = 6
    clipboard_record = 7
    initial_fill_rule_record = 8


class ImageResourceID(enum.IntEnum):
    """
    Ids for Image Resource blocks.
    """
    mac_print_info = 1001
    mac_page_format_info = 1002
    resolution_info = 1005
    alpha_channel_names = 1006
    caption = 1008
    border_info = 1009
    background_color = 1010
    print_flags = 1011
    grayscale_and_multichannel_halftoning_info = 1012
    color_halftoning_info = 1013
    duotone_halftoning_info = 1014
    grayscale_and_multichannel_transfer_func = 1015
    color_transfer_funcs = 1016
    duotone_transfer_funcs = 1017
    duotone_image_info = 1018
    effective_black_and_white = 1019
    eps_options = 1021
    quick_mask_info = 1022
    layer_state_info = 1024
    layers_group_info = 1026
    iptc_naa = 1028
    image_mode_for_raw = 1029
    jpeg_quality = 1030
    grid_and_guides_info = 1032
    copyright_flag = 1034
    url = 1035
    thumbnail_resource = 1036
    global_angle = 1037
    icc_profile = 1039
    watermark = 1040
    icc_untagged_profile = 1041
    effects_visible = 1042
    spot_halftone = 1043
    document_specific_ids_seed_number = 1044
    unicode_alpha_names = 1045
    indexed_color_table_count = 1046
    transparency_index = 1047
    global_altitude = 1049
    slices = 1050
    workflow_url = 1051
    jump_to_xpep = 1052
    alpha_identifiers = 1053
    url_list = 1054
    version_info = 1057
    exif_data_1 = 1058
    exif_data_2 = 1059
    xmp_metadata = 1060
    caption_digest = 1061
    print_scale = 1062
    pixel_aspect_ratio = 1064
    layer_comps = 1065
    alternate_duotone_colors = 1066
    alternate_spot_colors = 1067
    layer_selection_ids = 1069
    hdr_toning_info = 1070
    print_info = 1071
    layer_groups_enabled = 1072
    color_samplers_resource = 1073
    measurement_scale = 1074
    timeline_info = 1075
    sheet_disclosure = 1076
    display_info = 1077
    onion_skins = 1078
    count_info = 1080
    print_info_cs5 = 1082
    print_style = 1083
    mac_nsprintinfo = 1084
    win_devmode = 1085
    auto_save_file_path = 1086
    auto_save_format = 1087
    path_selection_state = 1088
    name_of_clipping_path = 2999
    origin_path_info = 3000
    image_ready_variables = 7000
    image_ready_data_sets = 7001
    image_ready_default_selected_state = 7002
    image_ready_7_rollover_expanded_state = 7003
    image_ready_rollover_expanded_state = 7004
    image_ready_save_layer_settings = 7005
    image_ready_version = 7006
    lightroom_workflow = 8000
    print_flags_info = 10000


class Units(enum.IntEnum):
    inches = 1
    cm = 2
    points = 3
    picas = 4
    columns = 5


class GuideDirection(enum.IntEnum):
    vertical = 0
    horizontal = 1


class PrintScaleStyle(enum.IntEnum):
    centered = 0
    size_to_fit = 1
    user_defined = 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools.extension import Extension
from Cython.Build import cythonize

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'cython',
    'numpy',
    'traitlets'
]

test_requirements = [
    'pytest'
]


extensions = [
    Extension(
        "pytoshop.packbits",
        ["pytoshop/packbits.pyx"]
    )
]

setup(
    name='pytoshop',
    version='0.3.0',
    description="A Python-based library to write Photoshop PSD files",
    long_description=readme + '\n\n' + history,
    author="Michael Droettboom",
    author_email='mdboom@gmail.com',
    url='https://github.com/mdboom/pytoshop',
    packages=[
        'pytoshop',
        'pytoshop.user'
    ],
    package_dir={'pytoshop':
                 'pytoshop'},
    include_package_data=True,
    install_requires=requirements,
    license="BSD license",
    zip_safe=False,
    keywords='pytoshop',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    ext_modules=cythonize(extensions)
)



# -*- coding: utf-8 -*-

"""
Documentation snippets.
"""

from __future__ import unicode_literals, absolute_import


read_single = """
Instantiate from a file-like object.

Parameters
----------
fd : file-like object
    Must be readable, seekable and open in binary mode.
"""


read = read_single + """
header : PsdFile object
    An object to get global file information from.
"""


write_single = """
Write to a file-like object.

Parameters
----------
fd : file-like object
    Must be writable, seekable and open in binary mode.
"""


write = write_single + """
header : PsdFile object
    An object to get global file information from.
"""


length = """
The length of the section, in bytes, not including its header.
"""


total_length = """
The length of the section, in bytes, including its header.
"""
